(* Proof of soundness and completeness of subtype inference
   for abstract types with variance and private abbreviations.
   Also prove a number of other properties of subtyping:
   symmetry, a form of transitivity, and monotony.
*)   

Require Import List Arith FSets FMaps Max Omega.
Set Implicit Arguments.

Definition name := nat.
Open Scope nat_scope.

Module Names.
  Definition t := name.
  Definition eq (x:t) y := x = y.
  Definition lt (x:t) (y:t) := x < y.
  Lemma eq_refl : forall x : t, eq x x.
  Proof. intros. unfold eq; reflexivity. Qed.
  Lemma eq_sym : forall x y : t, eq x y -> eq y x.
  Proof. unfold eq; intros; symmetry; auto. Qed.
  Lemma eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z.
  Proof. unfold eq; intros. rewrite H; auto. Qed.
  Lemma lt_trans : forall x y z : t, lt x y -> lt y z -> lt x z.
  Proof. unfold lt; intros. apply (lt_trans _ _ _ H H0). Qed.
  Lemma lt_not_eq : forall x y : t, lt x y -> ~ eq x y.
  Proof. unfold eq, lt; intros x y Hl He. rewrite He in Hl.
    apply (lt_irrefl _ Hl).
  Qed.
  Lemma compare : forall x y : t, Compare lt eq x y.
  Proof.
    intros.
    destruct (lt_eq_lt_dec x y).
      destruct s.
        apply LT. apply l.
      apply EQ; apply e.
    apply GT; apply l.
  Qed.
End Names.

Module NameSet := FSetList.Make(Names).
Definition names := NameSet.t.

Module Env := FMapList.Make(Names).
Definition env := Env.t.

Lemma math_ind : forall Q : nat -> Prop,
  (forall n, (forall m, m < n -> Q m) -> Q n) ->
  forall m, Q m.
Proof.
  intros.
  pose (n:= S m).
  assert (m < n). unfold n; omega.
  clearbody n.
  generalize dependent m.
  induction n; intros.
    assert False. omega. elim H1.
  apply H.
  intros; apply IHn.
  omega.
Qed.

Inductive polarity : Set := P | N.

Section Types.

Definition kind := list (bool * polarity).

Parameter variance : name -> option kind.

Inductive typ : Set :=
  | Tvar : name -> typ
  | Tconstr : name -> list typ -> typ.

Print typ_ind.
Reset typ_ind.

Fixpoint depth (t:typ) : nat :=
  match t with
  | Tvar _ => 0
  | Tconstr _ tl =>
      S (fold_right (fun t => max (depth t)) 0 tl)
  end.

Lemma typ_depth_ind : forall (P : typ -> Prop),
  (forall n,
    (forall t, depth t < n -> P t) -> (forall t, depth t <= n -> P t)) ->
  (forall t, P t).
Proof.
  intros.
  assert (depth t < S (depth t)). omega.
  set (n:=S(depth t)) in *. clearbody n.
  generalize dependent t.
  induction n; intros.
    elim (lt_n_O _ H0).
  apply (H n). auto.
  omega.
Qed.

Lemma typ_ind : forall (P : typ -> Prop),
  (forall n, P (Tvar n)) ->
  (forall n l, (forall x, In x l -> P x) -> P (Tconstr n l)) ->
  (forall t, P t).
Proof.
  intros.
  apply typ_depth_ind.
  intros.
  destruct t0. auto.
  apply H0.
  induction l; intros. elim H3.
  simpl in H3; destruct H3.
    subst.
    apply H1.
    apply lt_S_n.
    apply le_lt_n_Sm.
    eapply le_trans; try apply H2.
    simpl.
    apply le_n_S.
    apply le_max_l.
  apply IHl.
  eapply le_trans; try apply H2.
  simpl.
  apply le_n_S.
  apply le_max_r.
  auto.
Qed.

Inductive type : typ -> Prop :=
  | tvar : forall v, type (Tvar v)
  | tconstr : forall tc k tl,
      variance tc = Some k ->
      length k = length tl ->
      (forall t, In t tl -> type t) ->
      type (Tconstr tc tl).

Definition ctype := {t:typ | type t}.

Fixpoint subst (s : env typ) (t : typ) {struct t} : typ :=
  match t with
  | Tvar n =>
    match Env.find n s with
    | None => t
    | Some t' => t'
    end
  | Tconstr tc tl => Tconstr tc (map (subst s) tl)
  end.

Definition negp p :=
  match p with
  | P => N
  | N => P
  end.

Definition multp p1 p2 :=
  match p1 with
  | P => p2
  | N => negp p2
  end.

Parameter expansion : name -> option typ.

Fixpoint nth (A:Set) (n:nat) (l:list A) {struct l} : option A :=
  match l with
  | nil => None
  | a :: l' =>
    match n with
    | 0 => Some a
    | S n' => nth n' l'
    end
  end.

Definition andp (p p' : polarity) := if p then p' else N.

Definition mult2p b p b' p' :=
  match b, b' with
  | true,  true  => multp p p'
  | true,  false => p'
  | false, _  => p
  end.

Inductive For_all2(A B:Set)(P:A->B->Prop) : list A -> list B -> Prop :=
  | FA2nil : For_all2 P nil nil
  | FA2cons : forall a b l1 l2,
      P a b ->
      For_all2 P l1 l2 ->
      For_all2 P (a :: l1) (b :: l2).

Inductive inst_check(tc:name) : bool -> polarity -> typ -> Prop :=
  | ic_var1 : forall k p n,
      variance tc = Some k ->
      nth n k = Some(true, p) -> inst_check tc true p (Tvar n)
  | ic_var2P : forall k b p n,
      variance tc = Some k ->
      nth n k = Some(false, P) -> inst_check tc b p (Tvar n)
  | ic_var2N : forall k n,
      variance tc = Some k ->
      nth n k <> None -> inst_check tc false N (Tvar n)
  | ic_constr : forall b p n k' tl,
      n < tc ->
      variance n = Some k' ->
      length k' = length tl ->
      (forall b' p' t, In ((b',p'),t) (combine k' tl) ->
        inst_check tc (b && b') (mult2p b p b' p') t) ->
      inst_check tc b p (Tconstr n tl).

Parameter expansion_ok1 : forall n t,
  expansion n = Some t -> inst_check n true P t.

Parameter expansion_ok2 : forall n n' tl,
  expansion n = Some (Tconstr n' tl) -> n' < n.

Parameter expansion_ok3 : forall n n',
  not (expansion n = Some (Tvar n')).

Fixpoint expansion_depth (n : name) : nat :=
  match n with
  | 0 => 1
  | S m =>
    match expansion n with
    | None => 1
    | Some t => S (depth t)
    end * S (expansion_depth m)
  end.

Fixpoint typ_depth (t:typ) : nat :=
  match t with
  | Tvar _ => 0
  | Tconstr n tl =>
    expansion_depth n + fold_right (fun x => max (typ_depth x)) 0 tl
  end.

Lemma max_list_le : forall (A:Set) (x:A) f l a,
  In x l -> f x <= fold_right (fun x => max (f x)) a l.
Proof.
  induction l; intros. elim H.
  simpl in *; destruct H; subst. apply le_max_l.
  eapply le_trans. apply (IHl a0 H). apply le_max_r.
Qed.

Hint Resolve max_list_le.

Lemma expansion_depth_lt : forall n m,
  n < m -> expansion_depth n < expansion_depth m.
Proof.
  intros.
  pose (h := m - S n).
  assert (m = h + S n). unfold h. omega.
  clearbody h; subst. clear H.
  induction h; simpl.
    destruct (expansion (S n)); simpl; auto with arith.
  destruct (expansion (S (h+ S n))); simpl; auto with arith.
Qed.

Lemma expansion_depth_le : forall n m,
  n <= m -> expansion_depth n <= expansion_depth m.
Proof.
  intros.
  destruct (le_lt_or_eq _ _ H).
    apply lt_le_weak. apply (expansion_depth_lt H0).
  subst; auto.
Qed.

Lemma expansion_depth_nz : forall n, 1 <= expansion_depth n.
Proof.
  intros.
  assert (0 <= n) by auto with arith.
  apply (expansion_depth_le H).
Qed.

Lemma expansion_depth_nz2 : forall n,
  expansion_depth n = S (expansion_depth n - 1).
Proof.
  intros.
  pose (expansion_depth_nz n).
  omega.
Qed.

Fixpoint mkenv (n:nat) (tl:list typ) {struct tl} : env typ :=
  match tl with
  | nil => Env.empty typ
  | t::tl' => Env.add n t (mkenv (S n) tl')
  end.

Lemma subst_n : forall tl n,
  subst (mkenv 0 tl) (Tvar n) = List.nth n tl (Tvar n).
Proof.
  intros.
  pattern n at 2.
  replace n with (n-0) by omega.
  assert (0 <= n) by omega.
  generalize dependent n.
  set (k:=0). clearbody k.
  revert k.
  induction tl; intros.
    simpl. destruct (n-k); auto.
  case_eq (n-k); intros.
    pose (le_plus_minus _ _ H).
    rewrite H0 in e. rewrite plus_0_r in e.
    subst; clear H H0.
    simpl.
    rewrite (Env.find_1 (Env.add_1 (mkenv (S k) tl) a (refl_equal k))).
    auto.
  simpl.
  assert (S k <= n). omega.
  pose (IHtl _ _ H1). clearbody e. simpl in e.
  assert (k <> n). omega.
  assert (n - S k = n0). omega.
  subst.
  case_eq (Env.find n (mkenv (S k) tl)); intros.
    rewrite H3 in *.
    pose (Env.find_2 H3).
    rewrite (Env.find_1 (Env.add_2 a H2 m)).
    rewrite e. auto.
  case_eq (Env.find n (Env.add k a (mkenv (S k) tl))); intros.
    clear e.
    rewrite (Env.find_1 (Env.add_3 H2 (Env.find_2 H4))) in H3.
    discriminate.
  rewrite H3 in e. auto.
Qed.

Lemma in_combine_r' : forall (A B:Set) (x:B) (l1:list A) (l2:list B),
  In x l2 ->
  length l1 = length l2 ->
  exists y, In (y,x) (combine l1 l2).
Proof.
  induction l1; intros; destruct l2; try discriminate.
    elim H.
  simpl in H; destruct H.
    subst. esplit; simpl; eauto.
  destruct (IHl1 l2 H).
    simpl in H0; inversion H0. auto.
  exists x0; simpl; auto.
Qed.

Lemma expansion_lt : forall n tl t,
  expansion n = Some t ->
  typ_depth (subst (mkenv 0 tl) t) < typ_depth (Tconstr n tl).
Proof.
  intros.
  simpl.
  generalize (expansion_ok1 H); intro.
  destruct n; simpl.
    destruct t.
      rewrite subst_n.
      destruct (nth_in_or_default n tl (Tvar n)).
        assert (0 <= 0) by auto with arith.
        pose (expansion_depth_le H1).
        apply le_lt_n_Sm. auto.
      rewrite e. simpl; auto with arith.
    pose (expansion_ok2 H). elimtype False; omega.
  rewrite H.
  clear H; revert H0.
  generalize true P.
  induction t; intros.
    rewrite subst_n.
    simpl.
    apply le_lt_n_Sm.
    destruct (nth_in_or_default n0 tl (Tvar n0)).
       rewrite plus_0_r. rewrite plus_comm. auto with arith.
    rewrite e. simpl; auto with arith.
  simpl depth. simpl typ_depth.
  inversion H0; subst.
  pose (expansion_depth_le (lt_n_Sm_le _ _ H3)). clearbody l0.
  assert (forall t, In t l ->
    expansion_depth n0 + typ_depth (subst (mkenv 0 tl) t) <
    S (S (depth t)) * S (expansion_depth n) +
    fold_right (fun x : typ => max (typ_depth x)) 0 tl).
    intros.
    simpl.
    apply le_lt_n_Sm.
    rewrite <- plus_assoc.
    apply plus_le_compat. auto.
    apply lt_le_weak.
    destruct (in_combine_r' _ _ _ H1 H7).
    destruct x.
    eapply (H t H1).
    apply H8.
    apply H2.
  clear -H1 l0.
  induction l.
    simpl. apply le_lt_n_Sm.
    rewrite plus_0_r. rewrite <- plus_assoc. auto with arith.
  simpl fold_right.
  destruct (max_dec (typ_depth (subst (mkenv 0 tl) a))
             (fold_right (fun x : typ => max (typ_depth x)) 0
               (map (subst (mkenv 0 tl)) l)));
    rewrite e; clear e.
    eapply lt_le_trans. apply H1. simpl; auto.
    apply plus_le_compat_r.
    apply mult_le_compat_r.
    repeat apply lt_n_S.
    apply le_lt_n_Sm.
    apply le_max_l.
  eapply lt_le_trans; try apply IHl.
    intros; apply H1. simpl; auto.
  clear H1 IHl.
  apply plus_le_compat_r.
  apply mult_le_compat_r.
  repeat apply lt_n_S.
  apply le_lt_n_Sm.
  apply le_max_r.
Qed.

Lemma args_depth_lt : forall n t tl,
  In t tl -> typ_depth t < typ_depth (Tconstr n tl).
Proof.
  intros.
  simpl.
  eapply le_lt_trans. apply (max_list_le t typ_depth tl 0 H).
  eapply lt_le_trans; try apply (plus_le_compat_r 1).
    simpl; auto with arith.
  apply expansion_depth_nz.
Qed.

End Types.

Section Subtyping.

Definition unify s t1 t2 := subst s t1 = subst s t2.

Inductive nonvar : typ -> Prop :=
  nvconstr : forall n tl, nonvar (Tconstr n tl).

Inductive subtype : polarity -> env typ -> typ -> typ -> Prop :=
  | SEqual : forall t1 t2 s p,
      unify s t1 t2 ->
      subtype p s t1 t2
  | SConstr : forall n k tl1 p s tl2,
      variance n = Some k ->
      length k = length tl1 ->
      length tl1 = length tl2 ->
      (forall p' t1 t2,
        In ((true,p'),(t1,t2)) (combine k (combine tl1 tl2)) ->
        subtype (multp p p') s t1 t2) ->
      (forall t1 t2,
        In ((false,P),(t1,t2)) (combine k (combine tl1 tl2)) ->
        unify s t1 t2) ->
      subtype p s (Tconstr n tl1) (Tconstr n tl2)
  | SSubP : forall n k tl t t' s,
      variance n = Some k ->
      expansion n = Some t ->
      length k = length tl ->
      nonvar t' ->
      subtype P s (subst (mkenv 0 tl) t) t' ->
      subtype P s (Tconstr n tl) t'
  | SSubN : forall n k tl t t' s,
      variance n = Some k ->
      expansion n = Some t ->
      length k = length tl ->
      nonvar t' ->
      subtype N s t' (subst (mkenv 0 tl) t) ->
      subtype N s t' (Tconstr n tl).

Print subtype_ind.

Hint Constructors type subtype nonvar.

Definition unifier s E :=
  forall t1 t2, In (t1,t2) E -> unify s t1 t2.

Definition id := Env.empty typ.

Lemma unify_refl : forall s t, unify s t t.
Proof. intros; unfold unify; auto. Qed.

Hint Resolve unify_refl.

Theorem subtyp_refl : forall p t, subtype p id t t.
Proof. intros; auto. Qed.

Lemma unifier_nil: forall s, unifier s nil.
Proof.
  intros; intro; intros. elim H.
Qed.

Lemma unifier_hd : forall s t1 t2 E,
  unifier s ((t1,t2)::E) -> unify s t1 t2.
Proof.
  intros. apply H. simpl; auto.
Qed.

Lemma unifier_tl : forall s p E,
  unifier s (p::E) -> unifier s E.
Proof.
  intros; intro; intros.
  apply H. simpl; auto.
Qed.

Lemma unifier_app : forall s E1 E2,
  unifier s E1 -> unifier s E2 -> unifier s (E1 ++ E2).
Proof.
  intros; intro; intros.
  destruct (in_app_or E1 E2 _ H1); auto.
Qed.

Lemma unifier_cons : forall s t1 t2 E,
  unify s t1 t2 -> unifier s E -> unifier s ((t1,t2)::E).
Proof.
  intros; intro; intros.
  simpl in H1; destruct H1.
    inversion H1; subst; auto.
  auto.
Qed.

Hint Resolve unifier_nil unifier_tl unifier_app unifier_cons.

Lemma unify_instance :
  forall s tl1 tl2,
    map (subst s) tl1 = map (subst s) tl2 ->
    forall t,
      unify s (subst (mkenv 0 tl1) t) (subst (mkenv 0 tl2) t).
Proof.
  unfold unify.
  induction t.
    repeat rewrite subst_n.
    rewrite <- map_nth.
    rewrite <- (map_nth (subst s)).
    rewrite H; auto.
  simpl.
  apply (f_equal (Tconstr n)).
  induction l. simpl; auto.
  simpl.
  rewrite (H0 a); simpl; auto.
  rewrite IHl; simpl; auto.
  intros; apply H0.
  simpl; auto.
Qed.

Lemma nth_length : forall (A:Set) n (k:list A) x,
  nth n k = Some x -> n < length k.
Proof.
  induction n; destruct k; intros; try discriminate.
    simpl; omega.
  simpl in H.
  pose (IHn _ _ H).
  simpl; omega.
Qed.

Lemma nth_nth : forall (A:Set) n (k:list A) x d,
  nth n k = Some x -> List.nth n k d = x.
Proof.
  induction n; destruct k; simpl; intros; try discriminate.
    inversion H; auto.
  apply IHn ; auto.
Qed.

Lemma in_combine_nth : forall (k:kind) tl1 tl2 n a,
  length k = length tl1 ->
  length tl1 = length tl2 ->
  nth n k = Some a ->
  In (a, (subst (mkenv 0 tl1) (Tvar n), subst (mkenv 0 tl2) (Tvar n)))
     (combine k (combine tl1 tl2)).
Proof.
  intros.
  assert (0 <= n). omega.
  repeat (rewrite subst_n; auto).
  repeat rewrite <- minus_n_O.
  rewrite <- combine_nth; auto.
  rewrite <- (nth_nth _ _ (true, P) H1).
  rewrite <- combine_nth.
  apply nth_In.
  repeat rewrite combine_length.
  rewrite (Min.min_l (length tl1)).
  rewrite Min.min_l.
  apply (nth_length _ _ H1).
  omega. omega. 
  rewrite combine_length.
  rewrite Min.min_l. omega. omega.
Qed.

Lemma combine_map : forall (A B:Set) (f1 f2:A->B) l,
  combine (map f1 l) (map f2 l) = map (fun x => (f1 x, f2 x)) l.
Proof.
  induction l; auto.
  simpl; rewrite IHl; auto.
Qed.

Lemma in_combine_mapr_inv : forall (A B C:Set) (f:B->C) x y (k:list A) l,
  In (x,y) (combine k (map f l)) ->
  exists z, f z = y /\ In (x,z) (combine k l).
Proof.
  induction k; destruct l; simpl; intros; try solve [elim H].
  destruct H.
    inversion H; subst.
    exists b; auto.
  destruct (IHk l H) as [z [Hz Hl]].
  exists z; auto.
Qed.

Lemma subtyp_expansion1 : forall k tl1 tl2 n s b t p,
  length k = length tl1 ->
  length tl1 = length tl2 ->
  variance n = Some k ->
  forall p',
  (forall p0 t1 t2,
    In (true,p0,(t1,t2)) (combine k (combine tl1 tl2)) ->
    subtype (multp p' p0) s t1 t2) ->
  (forall t1 t2,
    In ((false,P),(t1,t2)) (combine k (combine tl1 tl2)) ->
    unify s t1 t2) ->
  inst_check n b p t ->
  (b = true ->
    subtype (multp p' p) s
      (subst (mkenv 0 tl1) t) (subst (mkenv 0 tl2) t)) /\
  (b = false -> p = P ->
    unify s (subst (mkenv 0 tl1) t) (subst (mkenv 0 tl2) t)).
Proof.
  induction 6; intros.
        split; intros; try discriminate.
        apply H2.
          apply in_combine_nth; auto.
        rewrite H4 in H1; inversion H1; subst; auto.
      rewrite H4 in H1; inversion H1; clear H1; subst.
      split; intros; subst.
        apply SEqual. apply H3. apply in_combine_nth; auto.
      apply H3. apply in_combine_nth; auto.
    intuition; discriminate.
  clear H2 H3.
  split; intros; subst; simpl subst.
    eapply SConstr. apply H5. rewrite map_length; auto.
      repeat rewrite map_length; auto.
      clear -H8.
      intros.
      rewrite combine_map in H.
      destruct (in_combine_mapr_inv _ _ _ _ _ H) as [z [Hz Hl]].
      destruct (H8 _ _ _ Hl) as [Hs _].
      inversion Hz; subst; clear H Hz.
      replace (multp (multp p' p) p'0)
        with (multp p' (mult2p true p true p'0)).
      auto.
      destruct p'; destruct p; destruct p'0; reflexivity.
    clear -H8.
    rewrite combine_map; intros.
    destruct (in_combine_mapr_inv _ _ _ _ _ H) as [z [Hz Hl]].
    destruct (H8 _ _ _ Hl) as [_ Hu].
    inversion Hz; subst; clear H Hz.
    apply Hu; destruct p; tauto.
  clear -H6 H8.
  unfold unify.
  simpl. apply (f_equal (Tconstr n0)).
  generalize dependent tl.
  induction k'; intros; destruct tl; try discriminate. reflexivity.
  simpl.
  destruct a.
  destruct (H8 b p t); simpl; intuition.
  rewrite H0.
  rewrite IHk'; auto.
  intros; apply H8. simpl; auto.
Qed.

Lemma subtyp_subn : forall s n2 tl2 n1 tl1,
  subtype P s (Tconstr n1 tl1) (Tconstr n2 tl2) ->
  n1 < n2 -> False.
Proof.
  induction n1 using math_ind; intros.
  inversion H0; subst.
      unfold unify in H2; inversion H2; clear H2; subst.
      omega.
    omega.
  destruct t. apply (expansion_ok3 H5).
  pose (expansion_ok2 H5).
  eapply (H _ l0). simpl in H9. apply H9.
  omega.
Qed.

Lemma in_commute2 : forall b p' (t1 t2:typ) (k:kind) tl1 tl2,
  length k = length tl1 ->
  length tl1 = length tl2 ->
  In (b, p', (t1, t2)) (combine k (combine tl2 tl1)) ->
  In (b, p', (t2, t1)) (combine k (combine tl1 tl2)).
Proof.
  induction k; destruct tl1; destruct tl2; intros; try discriminate.
    elim H1.
  simpl in H1; destruct H1. inversion H1; simpl; auto.
  simpl; right; auto.
Qed.

Hint Resolve in_commute2.

Lemma multp_negp : forall p p', multp (negp p) p' = negp (multp p p').
Proof.
  destruct p; destruct p'; simpl; auto.
Qed.

Lemma subtyp_sym : forall p s t1 t2,
  subtype p s t1 t2 ->
  subtype (negp p) s t2 t1.
Proof.
  induction 1; eauto.
    apply SEqual. unfold unify in *; auto.
  eapply SConstr. apply H. rewrite H1 in H0; apply H0. apply (sym_equal H1).
    intros. rewrite multp_negp; auto.
  unfold unify in *; intros; symmetry; auto.
Qed.

Lemma subtyp_expansion : forall n k p s tl1 tl2 t,
  variance n = Some k ->
  expansion n = Some t ->
  subtype p s (Tconstr n tl1) (Tconstr n tl2) ->
  subtype p s (subst (mkenv 0 tl1) t) (subst (mkenv 0 tl2) t).
Proof.
  intros.
  inversion H1; clear H1; subst.
      unfold unify in H2; simpl in H2.
      inversion H2; clear H2.
      apply SEqual. apply unify_instance; auto.
    rewrite H5 in H; inversion H; clear H; subst.
    pose (expansion_ok1 H0).
    destruct (subtyp_expansion1 _ _ H6 H9 H5 _ H10 H11 i).
    intuition.
    replace p with (multp p P); destruct p; simpl; auto.
   assert False.
     destruct t0. apply (expansion_ok3 H5).
     pose (expansion_ok2 H5).
     apply (subtyp_subn H11 l0).
    elim H1.
  assert False.
    destruct t0. apply (expansion_ok3 H5).
    pose (expansion_ok2 H5).
    apply (subtyp_subn (subtyp_sym H11) l0).
  elim H1.
Qed.

Lemma subtyp_trans1 : forall t2 t3 p s,
  (forall x, exists y, subst s (Tvar x) = Tvar y) ->
  subtype p s t2 t3 ->
  forall t1,
  unify s t1 t2 ->
  subtype p s t1 t3.
Proof.
  induction 2; intros.
     apply SEqual. unfold unify. rewrite H1. apply H0.
    (* SConstr *)
    clear H3.
    unfold unify in H6.
    destruct t1.
      destruct (H n0).
      rewrite H3 in H6; discriminate.
    simpl in H6.
    inversion H6; subst; clear H6.
    generalize (f_equal (length(A:=typ)) H8).
    repeat rewrite map_length; intro.
    assert (forall a t1 t2,
      In (a, (t1, t2)) (combine k (combine l tl2)) ->
      exists t : typ,
        unify s t1 t /\ In (a, (t, t2)) (combine k (combine tl1 tl2))).
      clear -H1 H2 H3 H8.
      intros.
      generalize dependent tl1; generalize dependent tl2;
        generalize dependent l.
      induction k; intros; destruct l; destruct tl1; destruct tl2;
        try discriminate; simpl in *.
        elim H.
      destruct H.
        inversion H; subst; clear H.
        exists t0.
        unfold unify; inversion H8; auto.
      inversion H1; clear H1.
      inversion H2; clear H2.
      inversion H3; clear H3.
      inversion H8; clear H8.
      destruct (IHk _ _ H tl1) as [t' [Hu Hi]]; clear IHk; auto.
      exists t'; eauto.
    eapply SConstr. apply H0. rewrite H3; auto. rewrite H3; auto.
      intros.
      destruct (H6 _ _ _ H7) as [t [Hu Ht]].
      apply (H4 _ _ _ Ht H _ Hu).
    intros.
    destruct (H6 _ _ _ H7) as [t [Hu Ht]].
    unfold unify; rewrite Hu.
    apply (H5 _ _ Ht).
    (* SSubP *)
    destruct t1.
      destruct (H n0).
      rewrite H5 in H6; discriminate.
    unfold unify in H5; simpl in H5.
    inversion H5; subst; clear H5.
    generalize (IHsubtype H (subst (mkenv 0 l) t)); intro; clear IHsubtype.
    generalize (f_equal (length(A:=typ)) H8).
    repeat rewrite map_length; intro.
    eapply SSubP; eauto. rewrite H6; auto.
    apply H5.
    apply unify_instance; auto.
    (* SSubN *)
    destruct t. elim (expansion_ok3 H1).
    eapply SSubN; eauto.
    inversion H3; subst; destruct t1.
      pose (H n2). clearbody e; destruct e. rewrite H5 in H6.
      discriminate.
    constructor.
Qed.

Lemma in_combine_mid : forall a t1 t3 (k:kind) (tl1 tl2 tl3:list typ),
  length tl1 = length tl2 ->
  length tl2 = length tl3 ->
  In (a, (t1, t3)) (combine k (combine tl1 tl3)) ->
  exists t2,
    In (a, (t1, t2)) (combine k (combine tl1 tl2)) /\
    In (a, (t2, t3)) (combine k (combine tl2 tl3)).
Proof.
  induction k; intros. elim H1.
  destruct tl1. elim H1.
  destruct tl2. discriminate.
  destruct tl3. elim H1.
  simpl in H1; destruct H1.
    inversion H1; subst; clear H1.
    exists t0; simpl; auto.
  simpl in H; inversion H; clear H.
  simpl in H0; inversion H0; clear H0.
  destruct (IHk _ _ _ H3 H2 H1).
  exists x; simpl; intuition.
Qed.

Lemma subtyp_trans1_sym : forall t2 t1 t3 p s,
  (forall x, exists y, subst s (Tvar x) = Tvar y) ->
  subtype p s t1 t2 ->
  unify s t2 t3 ->
  subtype p s t1 t3.
Proof.
  intros.
  replace p with (negp (negp p)) by (destruct p; reflexivity).
  apply subtyp_sym.
  eapply subtyp_trans1. auto.
    apply subtyp_sym. apply H0.
  unfold unify; symmetry; auto.
Qed.

Hint Resolve expansion_lt plus_lt_compat plus_lt_compat_l plus_lt_compat_r.

Lemma subtyp_trans : forall s t1 t2 t3 p,
  (forall x, exists y, subst s (Tvar x) = Tvar y) ->
  subtype p s t1 t2 ->
  subtype p s t2 t3 ->
  subtype p s t1 t3.
Proof.
  intros. generalize dependent p.
  remember (typ_depth t3 + typ_depth t1) as h.
  generalize dependent t3.
  revert t1 t2.
  induction h using math_ind; intros.
  assert (IHh: forall t2 t1 t3 p,
    typ_depth t3 + typ_depth t1 < h ->
    subtype p s t1 t2 -> subtype p s t2 t3 -> subtype p s t1 t3).
    intros. apply (H0 _ H3 t4 t0 t5); auto.
  clear H0.
  inversion H1; clear H1; subst.
  (* SEqual *)
  apply (subtyp_trans1 H H2); auto.
  (* SConstr *)
  inversion H2; clear H2; subst.
      apply (subtyp_trans1_sym (t2:=Tconstr n tl2)); auto.
      eauto.
    rewrite H8 in H0; inversion H0; subst; clear H0 H9.
    simpl; eapply SConstr; eauto. rewrite H4; auto.
      intros.
      destruct (in_combine_mid _ _ _ _ _ _ _ H4 H10 H0) as [t' [Hl Hr]].
      apply (IHh t' t1 t2); clear IHh; auto.
        apply plus_lt_compat; apply args_depth_lt.
          do 2 eapply in_combine_r. apply H0.
        eapply in_combine_l; eapply in_combine_r. apply H0.
      intros.
      destruct (in_combine_mid _ _ _ _ _ _ _ H4 H10 H0) as [t' [Hl Hr]].
      unfold unify. rewrite (H6 _ _ Hl). apply (H15 _ _ Hr).
    (* SSubP *)
    rewrite H8 in H0; inversion H0; subst; clear H0 H10.
    simpl; eapply SSubP; eauto.
    apply (IHh (subst (mkenv 0 tl2) t)); clear IHh; auto.
    eapply subtyp_expansion; eauto.
    (* SSubN *)
    eapply SSubN; eauto.
  (* SSubP *)
  eapply SSubP; eauto.
  inversion H5; subst; auto.
  inversion H2; subst; clear H2; auto.
  destruct t3.
    pose (H n1). clearbody e; destruct e. rewrite <- H1 in H2.
    discriminate.
  auto.
  (* SSubN *)
  inversion H2; subst; clear H2.
    eapply (subtyp_trans1_sym (t2:=Tconstr n tl)); eauto.
    rewrite H8 in H0; inversion H0; subst; clear H0 H9.
    eapply SSubN; eauto. rewrite H4; auto.
    apply (IHh (subst (mkenv 0 tl) t)); clear IHh; eauto.
    eapply subtyp_expansion; eauto.
  eapply SSubN; eauto.
Qed.

(* Composition of substitutions *)

Definition compose s1 s2 :=
  Env.fold (fun x t => Env.add x (subst s1 t)) s2 s1.

Lemma InA_rev(A:Set)(eqA:A->A->Prop) : forall x l,
  InA eqA x l -> InA eqA x (rev l).
Proof.
  induction 1.
    simpl. induction (rev l). simpl. auto.
    simpl. auto.
  simpl. induction (rev l); inversion IHInA; subst; simpl; auto.
Qed.

Lemma eq_key_elt_refl : forall (A:Set) (x:Env.key*A), Env.eq_key_elt x x.
Proof.
  intros; split. apply Names.eq_refl. reflexivity.
Qed.

Definition is_map (A B:Set) (eq:(A*B)->(A*B)->Prop) (l:list(A*B)) :=
  forall x y z, InA eq (x,y) l -> InA eq (x,z) l -> y = z.

Lemma elements_is_map : forall (A:Set) s,
  is_map (Env.eq_key_elt(elt:=A)) (Env.elements(elt:=A) s).
Proof.
  intros; intro; intros.
  pose (Env.find_1 (Env.elements_2 H)).
  rewrite (Env.find_1 (Env.elements_2 H0)) in e.
  inversion e; auto.
Qed.

Lemma InA_appl : forall (A:Set)(eq:A->A->Prop)(x:A) l1 l2,
  InA eq x l1 -> InA eq x (l1 ++ l2).
Proof.
  induction 1; simpl; auto.
Qed.

Lemma maps_compose : forall s1 s2 n t,
  Env.MapsTo n t s2 -> Env.MapsTo n (subst s1 t) (compose s1 s2).
Proof.
  intros.
  pose (InA_rev (Env.elements_1 H)).
  unfold compose.
  rewrite Env.fold_1.
  generalize i.
  clear H i.
  pose (elements_is_map s2).
  clearbody i.
  induction (Env.elements s2) using rev_ind; intros. inversion i0.
  rewrite fold_left_app; destruct x.
  rewrite distr_rev in i0; simpl in *.
  inversion i0; subst.
    destruct H0. simpl in *. subst.
    apply Env.add_1. auto.
  destruct (eq_nat_dec n k).
    assert (InA (Env.eq_key_elt (elt:=typ)) (k, t0) ((k, t0) :: rev l)).
      constructor. split; auto. unfold Names.eq; auto.
    pose (InA_rev i0).
    clearbody i1. simpl in i1.  rewrite rev_involutive in i1.
    pose (InA_rev H). simpl in i2.  rewrite rev_involutive in i2.
    subst. rewrite (i _ _ _ i2 i1).
    apply Env.add_1. apply Names.eq_refl.
  apply Env.add_2.
    unfold Env.E.eq.
    intro; elim n0. rewrite H; auto.
  apply IHl.
    intro; intros. apply (i x). apply InA_appl. apply H.
    apply InA_appl; auto.
  auto.
Qed.

Lemma maps_compose2 : forall s1 s2 n t,
  ~ Env.In n s2 -> (Env.MapsTo n t s1 <-> Env.MapsTo n t (compose s1 s2)).
Proof.
  intros.
  unfold compose.
  rewrite Env.fold_1.
  assert (forall x,
    ~InA (Env.eq_key_elt(elt:=typ)) (n,x) (rev (Env.elements s2))).
    intros x ne.
    elim H.
    exists x.
    apply Env.elements_2.
    rewrite <- (rev_involutive (Env.elements s2)).
    apply (InA_rev ne).
  clear H.
  induction (Env.elements s2) using rev_ind; simpl. intuition.
  rewrite fold_left_app; simpl.
  rewrite distr_rev in H0; simpl in H0.
  destruct IHl.
    intros y ne. elim (H0 y). auto.
  split; intro.
    apply Env.add_2.
    destruct x; simpl.
    intro; elim (H0 t0).
    rewrite H3. constructor. split; auto. unfold Names.eq; simpl; auto.
    apply (H H2).
  apply H1.
  destruct x; simpl in H2.
  eapply Env.add_3.
  intro; elim (H0 t0).
  rewrite <- H3. constructor. split; unfold Names.eq; reflexivity.
  apply H2.
Qed.

Lemma subst_compose : forall s1 s2 t,
  subst (compose s1 s2) t = subst s1 (subst s2 t).
Proof.
  induction t; simpl.
      case_eq (Env.find n s2); intros.
        pose (maps_compose s1 (Env.find_2 H)).
        rewrite (Env.find_1 m). auto.
      assert (~ Env.In n s2). intros [t Ht].
        rewrite (Env.find_1 Ht) in H. discriminate.
      case_eq (Env.find n s1); intros.
        pose (proj1 (maps_compose2 _  _ H0) (Env.find_2 H1)).
        rewrite (Env.find_1 m).
        simpl.
        rewrite H1. auto.
      simpl.
      rewrite H1.
      case_eq (Env.find n (compose s1 s2)); intros; auto.
      pose (proj2 (maps_compose2 _ _ H0) (Env.find_2 H2)).
      rewrite (Env.find_1 m) in H1. discriminate.
    auto.
  rewrite map_map.
  apply (f_equal (Tconstr n)).
  induction l; simpl; auto.
  rewrite H; simpl; auto.
  rewrite IHl; auto.
  intros; apply H; simpl; auto.
Qed.

Lemma map_combine : forall (A B:Set) (f1 f2:A->B) l1 l2,
  combine (map f1 l1) (map f2 l2) =
  map (fun x => (f1 (fst x), f2 (snd x))) (combine l1 l2).
Proof.
  induction l1; destruct l2; simpl; auto.
  rewrite IHl1; auto.
Qed.

Lemma subst_mkenv : forall n k tl b p t,
  variance n = Some k ->
  length k = length tl ->
  inst_check n b p t ->
  forall s,
  subst (mkenv 0 (map (subst s) tl)) t = subst s (subst (mkenv 0 tl) t).
Proof.
  induction 3; intros;
    try(rewrite H1 in H; inversion H; subst; clear H;
        do 2 rewrite subst_n; rewrite <- map_nth;
        apply nth_indep; rewrite map_length; rewrite <- H0).
      apply (nth_length _ _ H2).
      apply (nth_length _ _ H2).
    case_eq (nth n0 k); intros.
      apply (nth_length _ _ H).
    elim (H2 H).
  simpl.
  apply (f_equal (Tconstr n0)).
  rewrite map_map.
  clear -H3 H5.
  generalize dependent tl0; induction k'; destruct tl0; simpl; intros; auto.
    discriminate.
  destruct a.
  rewrite (H5 b p t); auto.
  inversion H3; clear H3.
  rewrite IHk'; clear IHk'; auto.
  intros. apply (H5 b' p' t0). auto.
Qed.

Lemma monotony : forall s' s p t1 t2,
  subtype p (compose s' s) t1 t2 ->
  subtype p s' (subst s t1) (subst s t2).
Proof.
  intros until s.
  remember (compose s' s) as s1.
  generalize dependent s1.
  induction 2; intros; subst.
      (* SEqual *)
      eapply SEqual; eauto.
      unfold unify. do 2 rewrite <- subst_compose. apply H.
    (* SConstr *)
    simpl.
    eapply SConstr; eauto; repeat rewrite map_length; auto.
      intros.
      rewrite map_combine in H5.
      destruct (in_combine_mapr_inv _ _ _ _ _ H5) as [[x1 x2] [E1 I1]].
      inversion E1; clear E1; subst.
      apply H3; auto.
    intros.
    rewrite map_combine in H5.
    destruct (in_combine_mapr_inv _ _ _ _ _ H5) as [[x1 x2] [E1 I1]].
    inversion E1; clear E1; subst.
    unfold unify. do 2 rewrite <- subst_compose. apply H4; auto.
  (* SSubP *)
  simpl.
  eapply SSubP; eauto; repeat rewrite map_length; auto.
    inversion H2; subst. simpl; auto.
  pose (expansion_ok1 H0).
  rewrite (subst_mkenv _ H H1 i).
  auto.
  (* SSubN *)
  simpl.
  eapply SSubN; eauto; repeat rewrite map_length; auto.
    inversion H2; subst. simpl; auto.
  pose (expansion_ok1 H0).
  rewrite (subst_mkenv _ H H1 i).
  auto.
Qed.

Definition app_option (A:Set) (l1 l2 : option (list A)) :=
  match l1, l2 with
  | Some l1, Some l2 => Some (l1 ++ l2)
  | _, _ => None
  end.

Definition subtype_aux (si:polarity->typ->typ->nat->option(list(typ*typ)))
  p h  :=
  fun (ktt:(bool*polarity)*(typ*typ)) E =>
    let (bp,tt) := ktt in let (b,p0) := bp in let (u1,u2) := tt in
    if b then app_option (si (multp p p0) u1 u2 h) E else
    if p0 then app_option (Some ((u1,u2)::nil)) E else E.

Fixpoint subtype_infn (p:polarity) (t1 t2:typ) (h:nat) {struct h}
  : option (list (typ*typ)) :=
  match h, t1, t2 with
  | S h', Tconstr n1 l1, Tconstr n2 l2 =>
    if eq_nat_dec n1 n2 then
      if eq_nat_dec (length l1) (length l2) then
        match variance n1 with
        | Some k =>
          if eq_nat_dec (length k) (length l1) then
          fold_right (subtype_aux subtype_infn p h')
            (Some nil) (combine k (combine l1 l2))
          else Some ((t1,t2)::nil)
        | None => Some ((t1,t2)::nil)
        end
      else None
    else match p with
    | P =>
      match expansion n1, variance n1 with
      | Some t, Some k =>
        if eq_nat_dec (length k) (length l1) then
          subtype_infn p (subst (mkenv 0 l1) t) (Tconstr n2 l2) h'
        else None
      | _, _ => None
      end
    | N =>
      match expansion n2, variance n2 with
      | Some t, Some k =>
        if eq_nat_dec (length k) (length l2) then
          subtype_infn p (Tconstr n1 l1) (subst (mkenv 0 l2) t) h'
        else None
      | _, _ => None
      end
    end
  | _, _, _ => Some ((t1,t2)::nil)
  end.

Lemma unifier_incl : forall s l2 l1,
  unifier s l2 -> incl l1 l2 -> unifier s l1.
Proof.
  induction l1; intros; intro; intros. elim H1.
  simpl in H1; destruct H1.
    subst.
    apply H. apply H0. simpl; auto.
  apply IHl1; auto.
  intro; intros. apply H0; simpl; auto.
Qed.
Hint Resolve incl_appl incl_appr incl_refl.

Lemma inference_sound : forall s h p t1 t2 E,
  subtype_infn p t1 t2 h = Some E -> unifier s E -> subtype p s t1 t2.
Proof.
  induction h; intros.
    simpl in H. inversion H; clear H; subst.
    apply SEqual. apply (unifier_hd H0).
  destruct t1; simpl in H.
    inversion H; clear H; subst.
    apply SEqual. apply (unifier_hd H0).
  destruct t2.
    inversion H; clear H; subst.
    apply SEqual. apply (unifier_hd H0).
  destruct (eq_nat_dec n n0).
    subst.
    destruct (eq_nat_dec (length l) (length l0)).
      case_eq (variance n0); intros.
        rewrite H1 in *.
        destruct (eq_nat_dec (length k) (length l)).
        assert (forall b p' t1 t2,
          In (b, p', (t1, t2)) (combine k (combine l l0)) ->
          (b = true -> subtype (multp p p') s t1 t2) /\
          (b = false /\ p' = P -> unify s t1 t2)).
        do 4 intro.
        generalize dependent E;
          generalize dependent l0; generalize dependent l.
        clear H1 n0.
        induction k; intros.
          elim H1.
        destruct l; destruct l0; try discriminate.
        simpl in H.
        simpl in e0; inversion e0; clear e0.
        simpl in e; inversion e; clear e.
        destruct a.
        destruct b0.
          case_eq (subtype_infn (multp p p0) t t0 h); intros;
            rewrite H2 in H;
            case_eq (fold_right (subtype_aux subtype_infn p h) 
              (Some nil) (combine k (combine l l0))); intros;
            rewrite H5 in H;
            try discriminate.
          simpl in H; inversion H; subst; clear H.
          simpl in H1; destruct H1.
            inversion H; subst; clear H IHk.
            split; intros.
              apply (IHh _ _ _ _ H2); auto. eapply unifier_incl; eauto.
            destruct H; discriminate.
          destruct (IHk _ H3 _ H4 l2); auto. eapply unifier_incl; eauto.
        destruct p0.
          case_eq (fold_right (subtype_aux subtype_infn p h) 
            (Some nil) (combine k (combine l l0))); intros; rewrite H2 in H.
          inversion H; clear H; subst.
          simpl in H1; destruct H1.
            inversion H; clear H; subst.
            split; intros. discriminate.
            apply (unifier_hd H0).
          apply (IHk _ H3 _ H4 l1); eauto.
         discriminate.
        simpl in H1; destruct H1.
          inversion H1; clear H1; subst. intuition; discriminate.
        apply (IHk _ H3 _ H4 E); auto.
        eapply SConstr; auto. apply H1. auto.
          intros; destruct (H2 true p' t1 t2); auto.
        intros; destruct (H2 false P t1 t2); auto.
      apply SEqual; eauto.
      inversion H; subst; apply (unifier_hd H0).
    rewrite H1 in H.
    inversion H; subst; apply SEqual; apply (unifier_hd H0).
   discriminate.
  destruct p.
    case_eq (expansion n); intros; rewrite H1 in H;
      case_eq (variance n); intros; try rewrite H2 in H;
        try destruct (eq_nat_dec (length k) (length l));
          try discriminate.
    eapply SSubP; eauto.
  case_eq (expansion n0); intros; rewrite H1 in H;
    case_eq (variance n0); intros; try rewrite H2 in H;
      try destruct (eq_nat_dec (length k) (length l0));
        try discriminate.
  eapply SSubN; eauto.
Qed.

Lemma inference_complete1 : forall n s t1 t2 p,
  typ_depth t1 + typ_depth t2 <= n ->
  unify s t1 t2 ->
  exists E, subtype_infn p t1 t2 n = Some E /\ unifier s E.
Proof.
  induction n; intros.
    destruct t1; destruct t2; simpl in *;
      try rewrite expansion_depth_nz2 in H;
      try elim (le_Sn_O _ H).
    esplit; split; try reflexivity; auto.
  destruct t1; simpl.
    esplit; split; try reflexivity; auto.
  destruct t2.
    esplit; split; try reflexivity; auto.
  destruct (eq_nat_dec n0 n1).
    subst.
    destruct (eq_nat_dec (length l) (length l0)).
      case_eq (variance n1); intros.
        destruct (eq_nat_dec (length k) (length l)).
          assert (forall t t0, In t l -> In t0 l0 ->
                    typ_depth t + typ_depth t0 <= n).
            intros. apply lt_n_Sm_le.
            eapply lt_le_trans; try apply H.
            apply plus_lt_compat; apply args_depth_lt; auto.
          clear H H1.
          unfold unify in H0; simpl in H0.
          inversion H0; clear H0.
          revert e0.
          generalize dependent l0; generalize dependent l.
          induction k; intros.
            simpl; esplit; split; try reflexivity; auto.
          destruct l. discriminate.
          destruct l0. discriminate.
          simpl in e; inversion e; clear e.
          simpl in e0; inversion e0; clear e0.
          simpl in H1; inversion H1; clear H1.
          destruct (IHk l l0); clear IHk; auto.
            intros; apply H2; simpl; auto.
          destruct H.
          simpl.
          destruct a.
          destruct b.
            destruct (IHn s t t0 (multp p p0)); clear IHn; auto.
              apply H2; simpl; auto.
            destruct H6. rewrite H6. rewrite H.
            simpl. exists (x0++x); auto.
          clear IHn.
          destruct p0.
            rewrite H.
            exists ((t,t0)::x); auto.
          rewrite H.
          exists x; auto.
        esplit; split; try reflexivity; auto.
      esplit; split; try reflexivity; auto.
    unfold unify in H0. simpl in H0.
    inversion H0; clear H0.
    generalize (f_equal (length(A:=typ)) H2); intro.
    do 2 rewrite map_length in H0.
    elim (n0 H0).
  elimtype False.
  unfold unify in H0. simpl in H0.
  inversion H0; clear H0.
  apply (n2 H2).
Qed.

Lemma inference_complete : forall n s t1 t2 p,
  typ_depth t1 + typ_depth t2 <= n ->
  subtype p s t1 t2 ->
  exists E, subtype_infn p t1 t2 n = Some E /\ unifier s E.
Proof.
  induction n; intros.
    destruct t1.
      inversion H0; subst.
        apply inference_complete1; auto.
      inversion H4.
    simpl in H. rewrite expansion_depth_nz2 in H. elim (le_Sn_O _ H).
  inversion H0; clear H0; subst.
        apply inference_complete1; auto.
      assert (forall t1 t2, In t1 tl1 -> In t2 tl2 ->
                typ_depth t1 + typ_depth t2 <= n).
        intros. apply lt_n_Sm_le.
        eapply lt_le_trans; try apply H.
        apply plus_lt_compat; apply args_depth_lt; auto.
      simpl.
      destruct (eq_nat_dec n0 n0); intuition.
      destruct (eq_nat_dec (length tl1) (length tl2)); intuition.
      clear e e0; rewrite H1.
      destruct (eq_nat_dec (length k) (length tl1)); intuition.
      clear e H H1; revert H2.
      generalize dependent tl1; generalize dependent tl2.
      induction k; intros.
        simpl. esplit; split; try reflexivity; auto.
      destruct tl1. discriminate.
      destruct tl2. discriminate.
      simpl in H2; inversion H2; clear H2.
      simpl in H3; inversion H3; clear H3.
      destruct (IHk tl2 tl1); clear IHk; auto.
        intros; apply H4; simpl; auto. 
        intros; apply H5; simpl; auto.
        intros; apply H0; simpl; auto.
      destruct H.
      simpl.
      destruct a.
      destruct b.
        destruct (IHn s t t0 (multp p p0)); clear IHn; auto.
          apply H0; simpl; auto.
          apply H4; simpl; auto.
        destruct H6. rewrite H6. rewrite H.
        simpl. exists (x0++x); auto.
      clear IHn.
      destruct p0.
        rewrite H.
        exists ((t,t0)::x); split; auto.
        apply unifier_cons; auto.
        apply H5; simpl; auto.
      rewrite H.
      exists x; auto.
    destruct (IHn s (subst (mkenv 0 tl) t) t2 P); clear IHn; auto.
      apply lt_n_Sm_le.
      eapply lt_le_trans; try apply H.
      auto.
    simpl.
    destruct t2. inversion H4.
    destruct (eq_nat_dec n0 n1).
      subst.
      destruct t.
        elim (expansion_ok3 H2).
      pose (expansion_ok2 H2).
      simpl in H5.
      elim (subtyp_subn H5 l1).
    rewrite H1; rewrite H2.
    destruct (eq_nat_dec (length k) (length tl)); intuition.
    rewrite H6. exists x; auto.
  destruct (IHn s t1 (subst (mkenv 0 tl) t) N); clear IHn; auto.
    apply lt_n_Sm_le.
    eapply lt_le_trans; try apply H. auto.
  simpl.
  destruct t1. inversion H4.
  destruct (eq_nat_dec n1 n0).
    subst.
    destruct t.
      elim (expansion_ok3 H2).
    pose (expansion_ok2 H2).
    simpl in H5.
    elim (subtyp_subn (subtyp_sym H5) l1).
  rewrite H1; rewrite H2.
  destruct (eq_nat_dec (length k) (length tl)); intuition.
  rewrite H6. exists x; auto.
Qed.

End Subtyping.