(* Subtyping3:
   subtyping with structural polymorphism and private abbreviations.
   Only proved soundness of inference *)

Set Implicit Arguments.
Require Import FSets Metatheory List Arith Max Omega.

Definition name := nat.
Open Scope nat_scope.

Module Names.
  Definition t := name.
  Definition eq (x:t) y := x = y.
  Definition lt (x:t) (y:t) := x < y.
  Lemma eq_refl : forall x : t, eq x x.
  Proof. intros. unfold eq; reflexivity. Qed.
  Lemma eq_sym : forall x y : t, eq x y -> eq y x.
  Proof. unfold eq; intros; symmetry; auto. Qed.
  Lemma eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z.
  Proof. unfold eq; intros. rewrite H; auto. Qed.
  Lemma lt_trans : forall x y z : t, lt x y -> lt y z -> lt x z.
  Proof. unfold lt; intros. apply (lt_trans _ _ _ H H0). Qed.
  Lemma lt_not_eq : forall x y : t, lt x y -> ~ eq x y.
  Proof. unfold eq, lt; intros x y Hl He. rewrite He in Hl.
    apply (lt_irrefl _ Hl).
  Qed.
  Lemma compare : forall x y : t, Compare lt eq x y.
  Proof.
    intros.
    destruct (lt_eq_lt_dec x y).
      destruct s.
        apply LT. apply l.
      apply EQ; apply e.
    apply GT; apply l.
  Qed.
End Names.

Module NameSet := FSetList.Make(Names).
Definition names := NameSet.t.

Lemma math_ind : forall Q : nat -> Prop,
  (forall n, (forall m, m < n -> Q m) -> Q n) ->
  forall m, Q m.
Proof.
  intros.
  pose (n:= S m).
  assert (m < n). unfold n; omega.
  clearbody n.
  generalize dependent m.
  induction n; intros.
    elimtype False; omega.
  apply H.
  intros; apply IHn.
  omega.
Qed.

Inductive polarity : Set := P | N.

Inductive typ : Set :=
  | Tvar : var -> typ
  | Tconstr : name -> list typ -> typ.

Reset typ_ind.

Fixpoint depth (t:typ) : nat :=
  match t with
  | Tvar _ => 0
  | Tconstr _ tl =>
      S (fold_right (fun t => max (depth t)) 0 tl)
  end.

Lemma typ_depth_ind : forall (P : typ -> Prop),
  (forall n,
    (forall t, depth t < n -> P t) -> (forall t, depth t <= n -> P t)) ->
  (forall t, P t).
Proof.
  intros.
  assert (depth t < S (depth t)) by auto with arith.
  set (n:=S(depth t)) in *. clearbody n.
  generalize dependent t.
  induction n; intros.
    elim (lt_n_O _ H0).
  apply (H n). auto.
  auto with arith.
Qed.
Print typ_depth_ind.

Hint Resolve le_max_l le_max_r le_n_S.

Lemma typ_ind : forall (P : typ -> Prop),
  (forall n, P (Tvar n)) ->
  (forall n l, (forall x, In x l -> P x) -> P (Tconstr n l)) ->
  (forall t, P t).
Proof.
  intros.
  apply typ_depth_ind; intros.
  destruct t0. auto.
  apply H0.
  induction l; intros. elim H3.
  simpl in H3; destruct H3.
    subst.
    apply H1.
    cut (S (depth x) <= n). auto.
    eapply le_trans; try apply H2.
    simpl; auto.
  apply IHl; auto.
  eapply le_trans; try apply H2.
  simpl; auto.
Qed.

(** Constraint domain *)

Module Cstr.
  Parameter cstr : Set.
  Parameter valid : cstr -> Prop.
  Parameter entails : cstr -> cstr -> Prop.
  Parameter entails_refl : forall c, entails c c.
  Parameter entails_trans : forall c1 c2 c3,
    entails c1 c2 -> entails c2 c3 -> entails c1 c3.
  Parameter unique : cstr -> list var.
  Parameter dominate : cstr -> cstr -> bool.
  Definition final c1 := forall c2, entails c2 c1 -> c1 = c2.
  Parameter final_dec : forall c1, {final c1}+{~final c1}.
  Hint Resolve entails_refl.
End Cstr.

Record ckind : Set := Kind {
  kind_cstr : Cstr.cstr;
  kind_rel  : list (var*typ) }.

Definition kind := option ckind.

Definition kenv := env kind.

Definition kinds := list (bool * polarity * var * kind).
(* (true, P) -> only covariant occurences
   (true, N) -> only contravariant occurences
   (false, P) -> both co- and contravariant occurences
   (false, N) -> no occurences *)

Parameter params : name -> option kinds.

Definition coherent k := forall x T U,
  In x (Cstr.unique (kind_cstr k)) ->
  In (x,T) (kind_rel k) -> In (x,U) (kind_rel k) -> T = U.

Definition entails K K' :=
  Cstr.entails (kind_cstr K) (kind_cstr K') /\
  forall T:var*typ, In T (kind_rel K') -> In T (kind_rel K).

Definition ckind_map f k :=
  match k with Kind kc kr =>
    Kind kc (List.map (fun XT:var*typ => (fst XT, f (snd XT))) kr)
  end.

Definition kind_map f K :=
  match K with
  | None => None
  | Some k => Some (ckind_map f k)
  end.

Fixpoint subst (S : env typ) (t : typ) {struct t} : typ :=
  match t with
  | Tvar n =>
    match get n S with
    | None => t
    | Some t' => t'
    end
  | Tconstr tc tl => Tconstr tc (map (subst S) tl)
  end.

Definition kind_open k Xs Vs := kind_map (subst (combine Xs Vs)) k.

Inductive well_kinded : kenv -> kind -> typ -> Prop :=
  | wk_any : forall K T,
      well_kinded K None T
  | wk_kind : forall K k x k',
      binds x (Some k') K ->
      entails k' k ->
      well_kinded K (Some k) (Tvar x).

Definition kinds_vars : kinds -> list var :=
  map (fun a:bool*polarity*var*kind => snd (fst a)).

Inductive type : kenv -> typ -> Prop :=
  | tvar : forall K v, v \in dom K -> type K (Tvar v)
  | tconstr : forall K tc ks tl,
      params tc = Some ks ->
      length ks = length tl ->
      (forall t, In t tl -> type K t) ->
      (forall a k t, In (a,k,t) (combine ks tl) ->
        well_kinded K (kind_open k (kinds_vars ks) tl) t) ->
      type K (Tconstr tc tl).

Definition list_snd (A B:Set) := map (@snd A B).

Definition list_fst (A B:Set) := map (@fst A B).

Definition kind_types k := list_snd (kind_rel k).

Definition kind_ok k := Cstr.valid (kind_cstr k) /\ coherent k.

Definition kenv_ok (K : kenv) :=
  ok K /\
  forall v k, binds v (Some k) K ->
    kind_ok k /\ list_forall (type K) (kind_types k).

Definition kenv_of_kinds : kinds -> kenv :=
  map (fun a:bool*polarity*var*kind => (snd (fst a), snd a)).

Parameter params_ok : forall n ks,
  params n = Some ks -> kenv_ok (kenv_of_kinds ks).

Definition negp p :=
  match p with
  | P => N
  | N => P
  end.

Definition multp p1 p2 :=
  match p1 with
  | P => p2
  | N => negp p2
  end.

Fixpoint nth (A:Set) (n:nat) (l:list A) {struct l} : option A :=
  match l with
  | nil => None
  | a :: l' =>
    match n with
    | 0 => Some a
    | S n' => nth n' l'
    end
  end.

Definition mult2p b p b' p' :=
  match b, b' with
  | true,  true  => multp p p'
  | true,  false => p'
  | false, _  => p
  end.

Inductive inst_check(tc:name)(ks:kinds) : bool -> polarity -> typ -> Prop :=
  | ic_var1 : forall k p x,
      In (true, p, x, k) ks -> inst_check tc ks true p (Tvar x)
  | ic_var2P : forall k b p x,
      In (false, P, x, k) ks -> inst_check tc ks b p (Tvar x)
  | ic_var2N : forall a x k,
      In (a,x,k) ks -> inst_check tc ks false N (Tvar x)
  | ic_constr : forall b p n tl,
      n < tc ->
      length ks = length tl ->
      (forall b' p' k' x t, In (b',p',x,k',t) (combine ks tl) ->
        inst_check tc ks (b && b') (mult2p b p b' p') t) ->
      inst_check tc ks b p (Tconstr n tl).

Parameter expansion : name -> option (typ * kinds).

Definition For_all (A:Set) (P:A->Prop) l := forall x, In x l -> P x.

Definition All_final (ks:kinds) :=
  forall p k, In (p, Some k) ks -> Cstr.final (kind_cstr k).

Parameter expansion_ok1 : forall n t ks1,
  expansion n = Some (t, ks1) ->
  exists ks2, params n = Some ks2
    /\ inst_check n (ks1++ks2) true P t
    /\ type (kenv_of_kinds (ks1++ks2)) t
    /\ All_final ks1
    /\ kenv_ok (kenv_of_kinds (ks1++ks2)).

Lemma expansion_ok2 : forall n n' tl ks,
  expansion n = Some (Tconstr n' tl, ks) -> n' < n.
Proof.
  intros.
  destruct (expansion_ok1 H) as [ks2 [_ [Inst _]]].
  inversion Inst; subst; try discriminate.
  auto.
Qed.

Parameter expansion_ok3 : forall n t0 ks1 ks2 b p x k l t,
  expansion n = Some (t0, ks1) ->
  params n = Some ks2 ->
  In (b, p, x, Some k) (ks1++ks2) ->
  In (l, t) (kind_rel k) ->
  inst_check n (ks1++ks2) b p t.

Definition unify s t1 t2 := subst s t1 = subst s t2.

Inductive nonvar (K:env kind) : typ -> Prop :=
  | nvconstr : forall n tl, nonvar K (Tconstr n tl)
  | nvvar : forall x k,
      binds x (Some k) K ->
      Cstr.final (kind_cstr k) -> nonvar K (Tvar x).

Inductive subtype(K:kenv)
  : polarity -> env typ -> list (var*var) -> typ -> typ -> Prop :=
  | SEqual : forall t1 t2 s R p,
      type K t1 ->
      type K t2 ->
      kenv_ok K ->
      unify s t1 t2 ->
      subtype K p s R t1 t2
  | SConstr : forall n k tl1 p s R tl2,
      params n = Some k ->
      length k = length tl1 ->
      length tl1 = length tl2 ->
      (forall p' x k' t1 t2,
        In ((true,p',x,k'),(t1,t2)) (combine k (combine tl1 tl2)) ->
        subtype K (multp p p') s R t1 t2) ->
      (forall x k' t1 t2,
        In ((false,P,x,k'),(t1,t2)) (combine k (combine tl1 tl2)) ->
        unify s t1 t2) ->
      subtype K p s R (Tconstr n tl1) (Tconstr n tl2)
  | SDom : forall cu c1 k1 c2 k2 (p:polarity) s R v1 v2,
      kenv_ok K ->
      (cu = if p then c2 else c1) ->
      binds v1 (Some (Kind c1 k1)) K ->
      binds v2 (Some (Kind c2 k2)) K ->
      Cstr.dominate cu (if p then c1 else c2) = true ->
      (forall l, In l (Cstr.unique cu) ->
        In l (list_fst k1) /\ In l (list_fst k2)) ->
      (forall l t1 t2, In l (Cstr.unique cu) ->
        In (l,t1) k1 -> In (l,t2) k2 ->
        subtype K p s ((if p then (v1,v2) else (v2,v1))::R) t1 t2) ->
      subtype K p s R (Tvar v1) (Tvar v2)
  | SRec : forall p s R v1 v2,
      (if (p:polarity) then In (v1,v2) R else In (v2,v1) R) ->
      subtype K p s R (Tvar v1) (Tvar v2)
  | SSubP : forall n t ks1 ks2 vl tl t' s R,
      expansion n = Some (t, ks1) ->
      params n = Some ks2 ->
      fresh (dom K) (length ks1) vl ->
      length ks2 = length tl ->
      nonvar K t' ->
      let S := combine (kinds_vars (ks1++ks2)) (map Tvar vl ++ tl) in
      subtype (K & combine vl (map (kind_map (subst S)) (list_snd ks1)))
        P s R (subst S t) t' ->
      subtype K P s R (Tconstr n tl) t'
  | SSubN : forall n t ks1 ks2 vl tl t' s R,
      expansion n = Some (t, ks1) ->
      params n = Some ks2 ->
      fresh (dom K) (length ks1) vl ->
      length ks2 = length tl ->
      nonvar K t' ->
      let S := combine (kinds_vars (ks1++ks2)) (map Tvar vl ++ tl) in
      subtype (K & combine vl (map (kind_map (subst S)) (list_snd ks1)))
        N s R t' (subst S t) ->
      subtype K N s R t' (Tconstr n tl).

Print subtype_ind.

Section Assoc.
Variables A B : Set.
Variable eq : forall a b:A, {a=b}+{a<>b}.

Definition assoc_dec : forall x (l:list(A*B)),
  {y | In (x,y) l} + {forall y, ~In (x,y) l}.
  induction l.
    right. intro; intro. elim H.
  destruct a.
  destruct (eq x a).
    rewrite e.
    left. exists b. left; reflexivity.
  destruct IHl.
    left. destruct s as [y Hy]. exists y. right; auto.
  right. simpl; intro; intro.
  destruct H.
    elim n. symmetry.
    apply (f_equal (fst(A:=A)(B:=B)) H).
  elim (n0 y). auto.
Defined.
Print assoc_dec.

Fixpoint assoc x (l:list(A*B)) {struct l} : option B :=
  match l with
  | nil => None
  | ab::l' => if eq (fst ab) x then Some (snd ab) else assoc x l'
  end.

Lemma assoc_in : forall x y l,
  assoc x l = Some y -> In (x,y) l.
Proof.
  induction l; intros. discriminate.
  simpl in H.
  destruct (eq (fst a) x).
    destruct a.
    simpl in *; inversion H; subst; auto.
  right. auto.
Qed.

Definition app_option (l1 l2 : option (list A)) :=
  match l2, l1 with
  | Some l2, Some l1 => Some (l1 ++ l2)
  | _, _ => None
  end.
End Assoc.

Definition subtype_aux (si:typ->typ->nat->option(list(typ*typ))) h  :=
  fun (ktt:(bool*polarity*var*kind)*(typ*typ)) E =>
    let (bpxk,tt) := ktt in let (bpx,k) := bpxk in
    let (b,p) := fst bpx in let (u1,u2) := tt in
    if b then app_option (if p then si u1 u2 h else si u2 u1 h) E else
    if p then app_option (Some ((u1,u2)::nil)) E else E.

Definition subtype_aux1 (si:typ->typ->nat->option(list(typ*typ))) l t2 h :=
  fun lt1 E =>
    if fst lt1 == l then app_option (si (snd lt1) t2 h) E else E.

Definition subtype_aux2  si k1 k2 h :=
  fun l E =>
    if assoc eq_var_dec l (kind_rel k1) then
      match assoc eq_var_dec l (kind_rel k2) with
      | Some t2 => fold_right (subtype_aux1 si l t2 h) E (kind_rel k1)
      | None => None
      end
    else None.

Fixpoint subtype_infn K (R:list(var*var)) (t1 t2:typ) (h:nat) {struct h}
  : option (list (typ*typ)) :=
  match h, t1, t2 with
  | S h', Tconstr n1 l1, Tconstr n2 l2 =>
    if eq_nat_dec n1 n2 then
      if eq_nat_dec (length l1) (length l2) then
        match params n1 with
        | Some k =>
          if eq_nat_dec (length k) (length l1) then
          fold_right (subtype_aux (subtype_infn K R) h')
            (Some nil) (combine k (combine l1 l2))
          else Some ((t1,t2)::nil)
        | None => Some ((t1,t2)::nil)
        end
      else None
    else match expansion n1, params n1 with
    | Some (t,ks1), Some ks2 =>
      if eq_nat_dec (length ks2) (length l1) then
        let vl := proj1_sig (var_freshes (dom K) (length ks1)) in
        let S := combine (kinds_vars (ks1++ks2)) (map Tvar vl ++ l1) in
        let K' := K & combine vl (map (kind_map (subst S)) (list_snd ks1)) in
        subtype_infn K' R (subst S t) (Tconstr n2 l2) h'
      else None
    | _, _ => None
    end
  | S h', Tvar v1, Tvar v2 =>
    match get v1 K, get v2 K with
    | Some (Some k1), Some (Some k2) =>
      if Cstr.dominate (kind_cstr k2) (kind_cstr k1) then
        fold_right (subtype_aux2 (subtype_infn K ((v1,v2)::R)) k1 k2 h')
          (Some nil) (Cstr.unique (kind_cstr k2))
      else Some ((t1,t2)::nil)
    | _, _ => Some ((t1,t2)::nil)
    end
  | _, _, _ =>
    Some ((t1,t2)::nil)
  end.

Hint Constructors type subtype nonvar.

Definition unifier s E :=
  forall t1 t2, In (t1,t2) E -> unify s t1 t2.

Definition id := empty (A:=typ).

Lemma unify_refl : forall s t, unify s t t.
Proof. intros; unfold unify; auto. Qed.

Hint Resolve unify_refl.

Theorem subtyp_refl : forall p K R t,
  kenv_ok K -> type K t -> subtype K p id R t t.
Proof. intros; auto. Qed.

Lemma unifier_nil: forall s, unifier s nil.
Proof.
  intros; intro; intros. elim H.
Qed.

Lemma unifier_hd : forall s t1 t2 E,
  unifier s ((t1,t2)::E) -> unify s t1 t2.
Proof. intuition. Qed.

Lemma unifier_tl : forall s p E,
  unifier s (p::E) -> unifier s E.
Proof.
  intros; intro. intuition.
Qed.

Lemma unifier_app : forall s E1 E2,
  unifier s E1 -> unifier s E2 -> unifier s (E1 ++ E2).
Proof.
  intros; intro; intros.
  destruct (in_app_or E1 E2 _ H1); auto.
Qed.

Lemma unifier_cons : forall s t1 t2 E,
  unify s t1 t2 -> unifier s E -> unifier s ((t1,t2)::E).
Proof.
  intros; intro; intros.
  simpl in H1; destruct H1.
    inversion H1; subst; auto.
  auto.
Qed.

Hint Resolve unifier_tl unifier_app unifier_cons.

Lemma in_commute2 : forall bpk (t1 t2:typ) (ks:kinds) tl1 tl2,
  length ks = length tl1 ->
  length tl1 = length tl2 ->
  In (bpk, (t1, t2)) (combine ks (combine tl2 tl1)) ->
  In (bpk, (t2, t1)) (combine ks (combine tl1 tl2)).
Proof.
  induction ks; destruct tl1; destruct tl2; intros; try discriminate.
    elim H1.
  simpl in H1; destruct H1. inversion H1; simpl; auto.
  simpl; auto.
Qed.

Hint Resolve in_commute2.

Lemma multp_negp : forall p p', multp (negp p) p' = negp (multp p p').
Proof.
  destruct p; destruct p'; simpl; auto.
Qed.

Lemma subtyp_sym : forall K p s R t1 t2,
  subtype K p s R t1 t2 ->
  subtype K (negp p) s R t2 t1.
Proof.
  induction 1; eauto.
    apply* SEqual. unfold unify in *; auto.
  apply* SConstr. congruence.
    intros. rewrite multp_negp; eauto.
  unfold unify in *; intros; symmetry. eauto.
  destruct p; simpl in *; subst; apply* SDom; intros; destruct* (H4 l).
  destruct* p.
Qed.

Lemma unifier_incl : forall s l2 l1,
  unifier s l2 -> incl l1 l2 -> unifier s l1.
Proof.
  induction l1; intros; intro; intros. elim H1.
  simpl in H1; destruct* H1.
Qed.
Hint Resolve incl_appl incl_appr incl_refl.

Ltac case_rewrite t H :=
  case_eq t; introv H; rewrite H in *; try discriminate.

Lemma subtype_aux1_some : forall si a t h l k E,
  fold_right (subtype_aux1 si a t h) (Some l) k = Some E ->
  exists l2, E = l2 ++ l.
Proof.
  induction k; intros.
    simpl in H. inversion H. exists (nil(A:=typ*typ)). simpl; auto.
  simpl in H. unfold subtype_aux1 at 1 in H.
  destruct a0; simpl in H.
  case_rewrite (fold_right (subtype_aux1 si a t h) (Some l) k) R1;
    destruct (v == a);
    destruct (si t0 t h); try discriminate;
    destruct* (IHk l0).
  inversion H; subst.
  esplit. rewrite* app_ass.
Qed.

Lemma fold_right_none : forall (A B:Set) (f:A->option B->option B) l,
  (forall x, f x None = None) ->
  fold_right f None l = None.
Proof.
  intros; induction l; simpl in *; auto.
  rewrite* IHl.
Qed.

Lemma subtype_aux1_none : forall si a t0 h k1,
  fold_right (subtype_aux1 si a t0 h) None k1 = None.
Proof.
  intros; apply fold_right_none.
  intros [n t].
  unfold subtype_aux1; simpl.
  destruct* (n == a).
Qed.

Lemma subtype_aux2_none : forall si k1 k2 h a,
  subtype_aux2 si k1 k2 h a None = None.
Proof.
  unfold subtype_aux2.
  intros; case_eq (assoc eq_var_dec a (kind_rel k1)); intros; auto.
  case_eq (assoc eq_var_dec a (kind_rel k2)); intros; auto.
  apply subtype_aux1_none.
Qed.

Lemma subtype_aux2_in : forall K R c1 k1 c2 k2 h ll E,
  fold_right (subtype_aux2 (subtype_infn K R) (Kind c1 k1) (Kind c2 k2) h)
    (Some nil) ll = Some E ->
  forall l : var, In l ll -> In l (list_fst k1) /\ In l (list_fst k2).
Proof.
  induction ll; intros E HS; intros. elim H.
  simpl in HS.
  unfold subtype_aux2 at 1 in HS.
  simpl in HS.
  case_rewrite (assoc eq_var_dec a k1) R0.
  case_rewrite (assoc eq_var_dec a k2) R1.
  simpl in H; destruct H.
    subst.
    clear -R0 R1.
    pose (in_map (@fst var typ) _ _ (assoc_in eq_var_dec _ _ R0)).
    pose (in_map (@fst var typ) _ _ (assoc_in eq_var_dec _ _ R1)).
    simpl in *. auto.
  case_rewrite (fold_right (subtype_aux2 (subtype_infn K R)
                  (Kind c1 k1) (Kind c2 k2) h) (Some nil) ll) R2.
    apply (IHll l0); auto.
  clear -HS.
  rewrite subtype_aux1_none in HS. discriminate.
Qed.

Lemma list_forall_in_map : forall (A B:Set) (f:A->B) (P:B->Prop) x l,
  In x l -> list_forall P (map f l) -> P (f x).
Proof.
  induction l; simpl; intros. elim H.
  inversions H0.
  destruct H; subst; auto.
Qed.

Lemma subtype_aux1_sound : forall K R t1 t2 h s l E0 k1 E,
  (forall E, subtype_infn K R t1 t2 h = Some E ->
             unifier s E -> subtype K P s R t1 t2) ->
  unifier s E ->
  kenv_ok K ->
  In (l, t1) k1 ->
  type K t1 ->
  type K t2 ->
  fold_right (subtype_aux1 (subtype_infn K R) l t2 h) (Some E0) k1 = Some E ->
  subtype K P s R t1 t2.
Proof.
  induction k1; introv; intros IHh HU HK Ht1 Tp1 Tp2 HS.
    elim Ht1.
  simpl in *.
  destruct a; unfold subtype_aux1 at 1 in HS; simpl in HS.
  destruct (v == l).
    subst.
    case_rewrite (fold_right
      (subtype_aux1 (subtype_infn K R) l t2 h) (Some E0) k1) R0.
    case_rewrite (subtype_infn K R t t2 h) R1.
    inversions HS; clear HS.
    destruct Ht1.
      inversions H; clear H; simpl in *.
      apply* IHh. apply* unifier_incl.
    apply* (IHk1 l0). apply* unifier_incl.
  destruct Ht1. inversions H; elim n; auto.
  apply* (IHk1 E).
Qed.

Lemma subtype_infn_dom : forall v v0 c1 k1 c2 k2 h l E K s R t1 t2,
  (forall R s t1 t2 E,
    subtype_infn K R t1 t2 h = Some E -> unifier s E ->
    kenv_ok K -> type K t1 -> type K t2 -> subtype K P s R t1 t2) ->
  binds v (Some (Kind c1 k1)) K ->
  binds v0 (Some (Kind c2 k2)) K ->
  fold_right
    (subtype_aux2 (subtype_infn K R) (Kind c1 k1) (Kind c2 k2) h)
    (Some nil) (Cstr.unique c2) = Some E ->
  unifier s E ->
  kenv_ok K ->
  In l (Cstr.unique c2) ->
  In (l, t1) k1 -> In (l, t2) k2 -> subtype K P s R t1 t2.
Proof.
  intros until t2; intros IHh Hv Hv0 HS HU HK H Ht1 Ht2.
  remember (Cstr.unique c2) as ll.
  generalize H.
  rewrite Heqll in H; clear Heqll.
  generalize dependent E; induction ll; intros; simpl in HS. elim H0.
  case_rewrite (fold_right
    (subtype_aux2 (subtype_infn K R) (Kind c1 k1) (Kind c2 k2) h)
    (Some nil) ll) R1.
    unfold subtype_aux2 in HS; simpl in HS.
    case_rewrite (assoc eq_var_dec a k1) R2.
    case_rewrite (assoc eq_var_dec a k2) R3.
    simpl in H0; destruct H0.
      subst.
      assert (t0 = t2).
        destruct* (proj2 HK v0 (Kind c2 k2)).
        apply (proj2 H0 l t0 t2); simpl; auto.
        apply (assoc_in eq_var_dec _ _ R3).
      subst.
      assert (Tp1: type K (snd (l,t1))).
        destruct (proj2 HK _ _ Hv) as [_ TT].
        eapply list_forall_in_map. apply Ht1. apply TT.
      assert (Tp2: type K (snd (l,t2))).
        destruct (proj2 HK _ _ Hv0) as [_ TT0].
        eapply list_forall_in_map. apply Ht2. apply TT0.
      simpl in *.
      eapply subtype_aux1_sound; try apply HS; auto*.
    apply (IHll l0); auto.
    destruct (subtype_aux1_some _ _ _ _ _ _ HS).
    subst; apply* unifier_incl.
  rewrite subtype_aux2_none in HS. discriminate.
Qed.

Lemma subtype_aux_some : forall si h l0 k l1,
  fold_right (subtype_aux si h) (Some l0) k = Some l1 ->
  exists l2, l2 ++ l0 = l1.
Proof.
  induction k; intros.
    inversion H. exists (nil(A:=typ*typ)). auto.
  simpl in H.
  unfold subtype_aux at 1 in H.
  destruct a as [[[[b p] x] k'][t1 t2]].
  destruct b; destruct p;
    case_rewrite (fold_right (subtype_aux si h) (Some l0) k) R0;
    destruct (IHk l); clear IHk; auto; subst.
        destruct (si t1 t2 h); inversion H.
        exists (l++x0); rewrite app_ass; auto.
      destruct (si t2 t1 h); inversion H.
      exists (l++x0); rewrite app_ass; auto.
    inversion H.
    exists ((t1,t2)::x0); simpl; auto.
  inversion H.
  exists x0; auto.
Qed.

Lemma subtype_infn_constr : forall K h,
  (forall K R s t1 t2 E,
    subtype_infn K R t1 t2 h = Some E -> unifier s E ->
    kenv_ok K -> type K t1 -> type K t2 -> subtype K P s R t1 t2) ->
  kenv_ok K ->
  forall R s k b p' x k' t1 t2,
  forall E l0 l,
    unifier s E ->
    length l = length l0 ->
    length k = length l ->
    list_forall (type K) l ->
    list_forall (type K) l0 ->
    fold_right (subtype_aux (subtype_infn K R) h) 
      (Some nil) (combine k (combine l l0)) = Some E ->
    In (b, p', x, k', (t1, t2)) (combine k (combine l l0)) ->
    (b = true -> subtype K p' s R t1 t2) /\
    (b = false /\ p' = P -> unify s t1 t2).
Proof.
  intros K h IHh.
  intros until t2.
  induction k; intros.
    elim H6.
  destruct l; destruct l0; try discriminate.
  simpl in *.
  inversion H1; clear H1.
  inversion H2; clear H2.
  destruct a.
  destruct p.
  destruct p. simpl in H5.
  inversions H3; inversions H4; clear H3 H4.
  destruct b0.
    case_rewrite (fold_right (subtype_aux (subtype_infn K R) h) 
      (Some nil) (combine k (combine l l0))) R1.
    destruct H6.
      inversions H1; clear H1 IHk.
      intuition; try discriminate.
      destruct p'.
        case_rewrite (subtype_infn K R t1 t2 h) R2.
        inversions H5; clear H5.
        assert (unifier s l2) by apply* unifier_incl.
        apply* IHh.
      case_rewrite (subtype_infn K R t2 t1 h) R2.
      inversions H5; clear H5.
      assert (unifier s l2) by apply* unifier_incl.
      replace N with (negp P) by reflexivity.
      apply subtyp_sym.
      apply* IHh. 
    apply (IHk l1 l0 l); clear IHk; auto.
    destruct
      (if p then subtype_infn K R t t0 h else subtype_infn K R t0 t h);
      inversion H5; subst.
    apply* unifier_incl.
  destruct H6.
    inversions H1; clear H1 IHk.
    intuition; try discriminate; subst.
    destruct (fold_right (subtype_aux (subtype_infn K R) h) 
            (Some nil) (combine k (combine l l0)));
      inversions H5.
    apply (unifier_hd H0).
  destruct p;
    case_rewrite (fold_right (subtype_aux (subtype_infn K R) h) 
                 (Some nil) (combine k (combine l l0))) R2;
    inversions H5; clear H5.
    apply (IHk l1 l0 l); clear IHk; auto.
    eauto.
  apply (IHk E l0 l); clear IHk; auto.
Qed.

Lemma notin_combine_fresh : forall (A:Set) Xs v (Vs:list A) n L,
  fresh L n Xs -> v \in L -> v # (combine Xs Vs).
Proof.
  induction Xs; simpl; intros. auto.
  destruct* Vs. auto with sets.
  destruct n. elim H.
  destruct H.
  apply (proj2 (notin_union v {{a}} (dom (combine Xs Vs)))).
  split.
    intro Hv. elim H.
    rewrite* <- (proj1 (in_singleton _ _) Hv).
  apply* IHXs. auto with sets.
Qed.

Lemma ok_cons : forall (A:Set) E x (a:A), ok E -> x # E -> ok ((x,a) :: E).
Proof. exact ok_push. Qed.

Lemma ok_combine_fresh : forall (A:Set) Xs Ks (Vs:list A),
  ok Ks -> fresh (dom Ks) (length Xs) Xs -> ok (Ks & combine Xs Vs).
Proof.
  induction Xs; destruct Vs; simpl; intros; auto*.
  apply* ok_cons.
  destruct H0.
  assert (a # combine Xs Vs). 
    apply* notin_combine_fresh. apply S.union_3. apply S.singleton_2.
    reflexivity.
  auto.
Qed.

Hint Resolve ok_combine_fresh.

Lemma well_kinded_weaken : forall K K' k t,
  well_kinded K k t -> ok (K & K') -> well_kinded (K & K') k t.
Proof.
  induction 1; intros; econstructor.
  apply* binds_concat_ok.
  auto.
Qed.

Lemma type_weaken : forall K K' t,
  type K t -> ok (K & K') -> type (K & K') t.
Proof.
  induction 1; intros.
    constructor.
    rewrite dom_concat. apply* S.union_2.
  apply* tconstr.
  intros.
  apply* well_kinded_weaken.
Qed.

Lemma binds_cons : forall (A:Set) x (a:A) E, binds x a ((x,a) :: E).
Proof. exact binds_head. Qed.

Lemma binds_subst_inv : forall v k1 vl S (k k0 : kinds),
  binds v k1 (combine vl (map (kind_map (subst S)) (list_snd k))) ->
  ok (kenv_of_kinds (k ++ k0)) ->
  length k = length vl ->
  exists x, exists k',
    binds x k' (kenv_of_kinds (k ++ k0)) /\ kind_map (subst S) k' = k1.
Proof.
  intros.
  unfold binds in H.
  gen vl; induction k; destruct vl; simpl; intros; try discriminate.
  destruct a. simpl in *.
  destruct (v == v0).
    subst. exists (snd p); exists k2.
    split. apply (binds_cons (snd p) k2).
    inversion* H.
  inversions H0.
  destruct* (IHk H4 vl) as [x [k' [B M]]].
  exists x; exists k'. split*.
  replace ((snd p, k2) :: kenv_of_kinds (k++k0))
    with (kenv_of_kinds (k++k0) & snd p ~ k2) by simpl*.
  auto.
Qed.

Lemma subst_in : forall v ks tl,
  v \in dom (kenv_of_kinds ks) ->
  length ks = length tl ->
  In (subst (combine (kinds_vars ks) tl) (Tvar v)) tl.
Proof.
  unfold kinds_vars.
  induction ks; destruct tl; simpl; intros; try discriminate.
    elim (in_empty H).
  destruct a as [[p x] k]. simpl in *.
  destruct* (v == x).
  inversion H0.
  destruct* (S.union_1 H). elim n. rewrite* (S.singleton_1 H1).
Qed.

Lemma combine_map_r : forall (A B C:Set)(f:B->C) (x:A) y l1 l2,
  In (x,y) (combine l1 (map f l2)) ->
  exists y', y = f y' /\ In (x,y') (combine l1 l2).
Proof.
  induction l1; destruct l2; simpl; intros; try contradiction.
  destruct H. inversion* H.
  destruct* (IHl1 l2 H).
Qed.

Lemma binds_combine_kinds : forall (ks:kinds) l x k,
  length ks = length l ->
  binds x k (kenv_of_kinds ks) ->
  exists a, exists t:typ, In (a,x,k,t) (combine ks l).
Proof.
  unfold binds.
  induction ks; destruct l; intros; try discriminate.
  destruct a as [[p v] k']; simpl in H0.
  destruct (x == v).
    subst. inversion* H0.
  inversion H.
  destruct* (IHks _ _ _ H2 H0) as [a [t0 HI]].
Qed.

Lemma entails_trans : forall k1 k2 k3,
  entails k1 k2 -> entails k2 k3 -> entails k1 k3.
Proof.
  intros.
  destruct H; destruct H0.
  split*.
  apply* (Cstr.entails_trans H H0).
Qed.

Lemma kind_subst_entails : forall S k k',
  entails k' k ->
  entails (ckind_map (subst S) k') (ckind_map (subst S) k).
Proof.
  intros.
  destruct H.
  destruct k as [kc kr]; destruct k' as [kc' kr'].
  split; simpl*.
  intros; simpl in *.
  destruct* (proj1 (in_map_iff _ _ _) H1) as [T' [e i]].
  subst.
  apply* (in_map (fun XT : var * typ => (fst XT, subst S (snd XT)))).
Qed.

Lemma env_map_combine : forall (A:Set) (f:A->A) (l1:list var) l2,
  length l1 = length l2 ->
  combine l1 (map f l2) = Env.map f (combine l1 l2).
Proof.
  induction l1; destruct l2; simpl; intros; try discriminate. auto.
  inversion H; rewrite* IHl1.
Qed.

Lemma dom_get : forall (A:Set) x (E:env A),
  x \in dom E -> get x E <> None.
Proof.
  induction E; simpl; intros. elim (in_empty H).
  destruct a. destruct (x == v). intro; discriminate.
  destruct* (S.union_1 H). elim n; symmetry; apply* S.singleton_1.
Qed.

Lemma list_map_ext : forall (A B:Set) (l:list A) (f1 f2:A->B),
  (forall x, In x l -> f1 x = f2 x) ->
  List.map f1 l = List.map f2 l.
Proof.
  intros. induction l. auto.
  simpl. rewrite* H. rewrite* IHl.
Qed.

Lemma subst_comm : forall K S1 S2 T,
  type K T ->
  dom K << dom S2 ->
  subst S1 (subst S2 T) = subst (Env.map (subst S1) S2) T.
Proof.
  induction 1; intros.
    simpl.
    case_eq (get v S2); intros.
      use (binds_map (subst S1) H1).
      unfold binds in H2; rewrite* H2.
    use (H0 _ H).
    elim (dom_get _ H2 H1).
  simpl.
  apply (f_equal (Tconstr tc)).
  rewrite map_map.
  apply list_map_ext. intros.
  apply* H2.
Qed.

Lemma kind_subst_comm : forall K S1 S2 k,
  list_forall (type K) (kind_types k) ->
  dom K << dom S2 ->
  ckind_map (fun T => subst S1 (subst S2 T)) k =
  ckind_map (subst (Env.map (subst S1) S2)) k.
Proof.
  intros.
  destruct k as [kc kr].
  unfold kind_types in H; simpl in *.
  apply (f_equal (Kind kc)).
  induction kr. auto.
  simpl in *. inversions H.
  rewrite* (subst_comm (K:=K)).
  rewrite* IHkr.
Qed.

Lemma subst_weaken : forall K S1 S2 T,
  type K T ->
  dom K << dom S1 ->
  (forall x k, binds x k S1 -> binds x k S2) ->
  subst S1 T = subst S2 T.
Proof.
  induction 1; intros.
    simpl.
    case_eq (get v S1); intros.
      unfold binds in H1. rewrite* (H1 _ _ H2).
    use (H0 _ H).
    elim (dom_get _ H3 H2).
  simpl.
  apply (f_equal (Tconstr tc)).
  apply list_map_ext. intros.
  apply* H2.
Qed.

Lemma kind_subst_weaken : forall K S1 S2 k,
  list_forall (type K) (kind_types k) ->
  dom K << dom S1 ->
  (forall x k, binds x k S1 -> binds x k S2) ->
  ckind_map (subst S1) k = ckind_map (subst S2) k.
Proof.
  intros.
  destruct k as [kc kr].
  unfold kind_types in H; simpl in *.
  apply (f_equal (Kind kc)).
  apply list_map_ext; intros.
  unfold list_snd in H.
  use (list_forall_in_map _ _ _ H2 H).
  rewrite* (subst_weaken (S2:=S2) H3 H0).
Qed.

Lemma ckind_map_map : forall f g k,
  ckind_map f (ckind_map g k) = ckind_map (fun T => f (g T)) k.
Proof.
  destruct k as [kc kr].
  simpl; rewrite* map_map.
Qed.

Lemma binds_in : forall (A:Set) x (a:A) E,
  In (x,a) E -> ok E -> binds x a E.
Proof.
  unfold binds.
  induction E; simpl; intros. elim H.
  destruct H.
    subst. destruct* (x==x).
  inversions H0.
  destruct* (x == x0).
  subst. elim (binds_fresh (IHE H H3) H4).
Qed.

Lemma dom_kenv_combine : forall ks (tl:list typ),
  length ks = length tl ->
  dom (kenv_of_kinds ks) = dom (combine (kinds_vars ks) tl).
Proof.
  unfold kenv_of_kinds, kinds_vars.
  induction ks; destruct tl; intros; try discriminate.
    auto.
  simpl.
  inversion H. rewrite* (IHks tl).
Qed.

Lemma in_dom_combine : forall (A:Set) x (K:env A) vl ks,
  In x vl ->
  length vl = length ks ->
  x \in dom (K & combine vl ks).
Proof.
  induction vl; destruct ks; intros; try discriminate. elim H.
  simpl in H; destruct H.
  subst. simpl. apply S.union_2. apply S.singleton_2. reflexivity.
  inversion H0. simpl.
  apply* S.union_3.
Qed.

Lemma combine_concat : forall (A:Set) l3 (l4:list A) l1 l2,
  length l1 = length l2 ->
  combine (l1 ++ l3) (l2 ++ l4) = combine l3 l4 & combine l1 l2.
Proof.
  induction l1; destruct l2; intros; try discriminate; auto.
  simpl. inversion H; rewrite* IHl1.
Qed.

Lemma ok_kinds_vars : forall (A:Set) ks (l:list A),
  ok (kenv_of_kinds ks) -> ok (combine (kinds_vars ks) l).
Proof.
  unfold kenv_of_kinds, kinds_vars; induction ks; simpl; intros.
    fold (empty(A:=A)). auto.
  destruct l.
    fold (empty(A:=A)). auto.
  inversions H; clear H.
  destruct a as [[a b] c]; simpl in *.
  apply* ok_cons.
  clear -H4.
  gen l; induction ks; simpl in *; intros. auto.
  destruct l. simpl. auto.
  simpl.
  destruct (proj1 (notin_union _ _ _) H4); clear H4.
  use (IHks H0 l).
Qed.

Lemma kinds_vars_app : forall ks1 ks2,
  kinds_vars (ks1 ++ ks2) = kinds_vars ks1 ++ kinds_vars ks2.
Proof.
  unfold kinds_vars; intros. rewrite* map_app.
Qed.

Lemma in_combine_map_l : forall (A B C:Set) (f:A->B) x (a:C) l1 l2,
  In (x,a) (combine l1 l2) ->
  In (f x, a) (combine (map f l1) l2).
Proof.
  induction l1; destruct l2; intros; try contradiction.
  simpl in H; destruct* H.
  inversion* H.
Qed.

Lemma binds_vars_prepend : forall K x a ks vl,
  binds x a (kenv_of_kinds ks) ->
  length ks = length vl ->
  exists v, binds x (Tvar v) (K & combine (kinds_vars ks) (map Tvar vl)).
Proof.
  induction ks; destruct vl; unfold binds; simpl; intros; try discriminate.
  destruct* (x == snd (fst a0)).
  inversion H; apply* (IHks vl).
Qed.

Lemma well_kinded_subst1 : forall K k vl n k0 l tc ks0 tl a x k' c ks,
  fresh (dom K) (length k) vl ->
  params n = Some k0 ->
  type K (Tconstr n l) ->
  ok K ->
  params tc = Some ks0 ->
  length ks0 = length tl ->
  kenv_ok (kenv_of_kinds (k ++ k0)) ->
  In (a, Some c, Tvar x) (combine ks0 tl) ->
  entails k' (ckind_map (subst (combine (kinds_vars ks0) tl)) c) ->
  let S := combine (kinds_vars (k ++ k0)) (map Tvar vl ++ l) in
  ok S ->
  length (kinds_vars k) = length (map Tvar vl) ->
  binds x (Some k') (kenv_of_kinds k0) ->
  well_kinded (K & combine vl ks)
    (kind_open (Some c) (kinds_vars ks0) (map (subst S) tl))
    (subst S (Tvar x)).
Proof.
  intros until ks. intros Fr Pn Tn OkK Ptc Len0 Kok HI Ent S OkS Len1 B.
  inversions Tn; clear Tn.
  rewrite H1 in Pn; inversions Pn; clear Pn.
  destruct (binds_combine_kinds _ _ H2 B) as [b [t' HI']].
  use (H5 _ _ _ HI'); clear H5.
  inversions H.
  assert (BxS: binds x (Tvar x0) S).
    apply* binds_in.
    unfold S.
    rewrite kinds_vars_app.
    rewrite* combine_concat.
    unfold concat.
    apply in_or_app.
    right.
    clear -HI'.
    use (in_combine_map_l (fun a:bool*polarity*var*kind => snd(fst a))
           _ _ _ _ HI').
  unfold binds in BxS.
  simpl; rewrite BxS.
  econstructor. auto*.
  apply* entails_trans.
  use (params_ok Ptc).
  use (params_ok H1).
  clear Ptc H1 H6 OkK.
  rewrite (kind_subst_weaken (K:=kenv_of_kinds k0) (S2:=S)
             (S1:=combine (kinds_vars k0) l) k').
        rewrite env_map_combine.
          rewrite <- (kind_subst_comm (K:=kenv_of_kinds ks0)).
              rewrite* <- ckind_map_map.
              apply* kind_subst_entails.
            destruct a.
            assert (In (v,Some c) (kenv_of_kinds ks0)).
              use (in_combine_l _ _ _ _ HI).
              use (in_map (fun a:bool*polarity*var*kind => (snd (fst a), snd a))
                     _ _ H1).
            use (binds_in _ _ H1 (proj1 H0)).
            apply (proj2 (proj2 H0 _ _ H6)).
          rewrite* <- dom_kenv_combine; apply subset_refl. 
        unfold kinds_vars. rewrite* map_length.
      apply (proj2 (proj2 H5 _ _ B)).
    rewrite* <- dom_kenv_combine; apply subset_refl. 
  intros.
  gen OkS.
  unfold S.
  rewrite kinds_vars_app.
  rewrite* combine_concat.
Qed.

Lemma well_kinded_subst2 : forall K k vl k0 l tc ks0 tl a c x k',
  fresh (dom K) (length k) vl ->
  ok K ->
  let S := combine (kinds_vars (k ++ k0)) (map Tvar vl ++ l) in
  let ks := map (kind_map (subst S)) (list_snd k) in
  params tc = Some ks0 ->
  length ks0 = length tl ->
  kenv_ok (kenv_of_kinds (k ++ k0)) ->
  In (a, Some c, Tvar x) (combine ks0 tl) ->
  entails k' (ckind_map (subst (combine (kinds_vars ks0) tl)) c) ->
  ok S ->
  length (kinds_vars k) = length (map Tvar vl) ->
  binds x (Some k') (kenv_of_kinds k) ->
  well_kinded (K & combine vl ks)
    (kind_open (Some c) (kinds_vars ks0) (map (subst S) tl))
    (subst S (Tvar x)).
Proof.
  introv Fr OkK. intros S ks. intros Ptc Len0 Kok HI Ent OkS Len1 B0.
  poses Len2 Len1.
  unfold kinds_vars in Len2. rewrite map_length in Len2.
  destruct (binds_combine_kinds _ _ Len2 B0) as [b [t' HI']].
  destruct (proj1 (in_map_iff _ _ _) (in_combine_r _ _ _ _ HI'))
    as [v [Ht' It']].
  subst.
  assert (BxS: binds x (Tvar v) S).
    apply* binds_in.
    unfold S.
    rewrite kinds_vars_app.
    rewrite* combine_concat.
    unfold concat.
    apply in_or_app.
    left.
    use (in_combine_map_l (fun a:bool*polarity*var*kind => snd(fst a))
           _ _ _ _ HI').
  unfold binds in BxS.
  simpl; rewrite BxS.
  econstructor.
    apply binds_prepend.
    cut (binds v (kind_map (subst S) (Some k')) (combine vl ks)).
      simpl; intro B; apply B.
    apply binds_in.
      unfold ks.
      clear -HI'.
      clearbody S.
      gen vl; induction k; intros.  elim HI'.
      destruct vl; simpl in *. elim HI'.
      destruct a; destruct (split k).
      destruct HI'.
        inversions H. simpl. auto.
      simpl in *. auto.
    rewrite app_nil_end.
    apply (ok_combine_fresh vl ks (Ks:=nil)).
      fold (empty(A:=option ckind)); auto.
    simpl*.
  rewrite env_map_combine.
    rewrite <- (kind_subst_comm (K:=kenv_of_kinds ks0)).
        rewrite* <- ckind_map_map.
        apply* kind_subst_entails.
      use (params_ok Ptc).
      destruct a.
      assert (Bc: binds v0 (Some c) (kenv_of_kinds ks0)).
        destruct H.
        apply* binds_in.
        use (in_combine_l _ _ _ _ HI).
        use (in_map (fun a:bool*polarity*var*kind => (snd (fst a), snd a))
               _ _ H1).
      apply (proj2 (proj2 H _ _ Bc)).
    rewrite* <- dom_kenv_combine. apply subset_refl.
  unfold kinds_vars; rewrite* map_length.
Qed.

Lemma type_subst : forall K (k k0:kinds) vl n l T,
  fresh (dom K) (length k) vl ->
  params n = Some k0 ->
  type K (Tconstr n l) ->
  ok K ->
  let S := combine (kinds_vars (k ++ k0)) (map Tvar vl ++ l) in
  let ks := map (kind_map (subst S)) (list_snd k) in
  type (kenv_of_kinds (k++k0)) T ->
  kenv_ok (kenv_of_kinds (k++k0)) ->
  type (K & combine vl ks) (subst S T).
Proof.
  intros until T. gen_eq (kenv_of_kinds (k++k0)) as K0.
  induction 6; intros; subst.
    assert (length ks = length k).
      unfold ks, list_snd; repeat rewrite* map_length.
    clearbody ks.
    assert (length (k++k0) = length (map Tvar vl ++ l)).
      do 2 rewrite app_length. rewrite map_length.
      rewrite (fresh_length _ _ _ H0).
      inversions H2. rewrite H8 in H1; inversions H1.
      rewrite* H9.
    destruct (in_app_or _ _ _ (subst_in _ _ H4 H6)); fold S in H7.
      destruct (proj1 (in_map_iff _ _ _) H7).
      destruct H8; rewrite <- H8.
      constructor.
      apply* in_dom_combine.
      rewrite H; rewrite* (fresh_length _ _ _ H0).
    rewrite <- H in H0.
    inversions H2.
    apply* type_weaken.
  simpl; econstructor; auto*. rewrite* map_length.
    intros. destruct (proj1 (in_map_iff _ _ _) H) as [x [HS HI]].
    subst; auto*.
  intros.
  destruct* (combine_map_r _ _ _ _ _ H) as [t0 [HE HI]].
  subst.
  use (H8 _ _ _ HI).
  clear -H10 H0 H1 H2 H3 H4 H5 H9 HI.
  inversions H10; clear H10; destruct k1; try discriminate.
    constructor.
  simpl in H.
  inversions H; clear H.
  replace (kenv_of_kinds (k++k0)) with (kenv_of_kinds k0 & kenv_of_kinds k)
    in H6 by (unfold concat; unfold kenv_of_kinds; rewrite* map_app).
  assert (OkS: ok S).
    unfold S.
    apply ok_kinds_vars.
    destruct* H9.
  assert (Len1: length (kinds_vars k) = length (map Tvar vl)).
    unfold kinds_vars. repeat rewrite map_length. apply* fresh_length.
  unfold ks, S.
  binds_cases H6.
    apply* well_kinded_subst1.
  apply* well_kinded_subst2.
Qed.

Lemma kind_ok_map : forall (f:typ->typ) kc kr,
  kind_ok (Kind kc kr) ->
  kind_ok (Kind kc  (map (fun XT : var * typ => (fst XT, f (snd XT))) kr)).
Proof.
  intros.
  destruct H; split*.
  intro; intros; simpl in *.
  destruct (proj1 (in_map_iff _ _ _) H2) as [[x1 T1] [HX1 HI1]].
  inversions HX1.
  destruct (proj1 (in_map_iff _ _ _) H3) as [[x2 T2] [HX2 HI2]].
  inversions HX2.
  rewrite* (H0 _ _ _ H1 HI1 HI2).
Qed.

Lemma kenv_ok_expansion : forall K n t k k0 l vl,
  kenv_ok K ->
  type K (Tconstr n l) ->
  expansion n = Some (t, k) ->
  params n = Some k0 ->
  fresh (dom K) (length k) vl ->
  let S := combine (kinds_vars (k ++ k0)) (map Tvar vl ++ l) in
  let ks := map (kind_map (subst S)) (list_snd k) in
  kenv_ok (K & combine vl ks).
Proof.
  intros until vl; intros HK HTp En Pn Fr S ks.
  poses OkK (proj1 HK).
  split*.
  intros.
  binds_cases H.
    destruct (proj2 HK _ _ B).
    split*.
    induction H0; auto.
    econstructor. auto. apply* type_weaken.
  destruct (expansion_ok1 En) as [ks2 [HP [_ [HT [_ Hok]]]]].
  rewrite HP in Pn. inversion Pn; clear Pn. subst ks2.
  use (binds_subst_inv _ _ _ _ B0 (proj1 Hok) (fresh_length _ _ _ Fr)).
  destruct H as [x [k' [B Hk1]]].
  destruct k' as [[kc kr]|]; try discriminate.
  inversions Hk1; clear Hk1.
  destruct (proj2 Hok _ _ B).
  simpl; split.
    apply* kind_ok_map.
  clear B B0 H.
  unfold kind_types in *.
  simpl kind_rel in *.
  induction kr. simpl*.
  simpl in *.
  inversions H0.
  constructor. auto.
  unfold ks, S; apply* type_subst.
Qed.

Lemma inference_sound : forall h K R s t1 t2 E,
  subtype_infn K R t1 t2 h = Some E ->
  unifier s E ->
  kenv_ok K ->
  type K t1 ->
  type K t2 ->
  subtype K P s R t1 t2.
Proof.
  induction h; intros until E; intros HS HU HK Ht1 Ht2.
    inversions HS.
    apply* SEqual.
  assert (Def: Some ((t1,t2)::nil) = Some E -> subtype K P s R t1 t2).
    intro H; inversions H. apply* SEqual.
  destruct t1; simpl in HS.
    destruct t2; auto.
    case_rewrite (get v K) R1; auto.
    destruct* o.
    case_rewrite (get v0 K) R2; auto.
    destruct* o.
    case_rewrite (Cstr.dominate (kind_cstr c0) (kind_cstr c)) R3; auto.
    destruct c0 as [c2 k2].
    destruct c as [c1 k1].
    simpl in *.
    apply* SDom; unfold binds; auto*.
      intros; apply (subtype_aux2_in _ _ _ _ _ _ _ _ HS). subst; auto.
    intros; apply* subtype_infn_dom; unfold binds; auto*.
  destruct t2; auto.
  destruct (n === n0).
    subst.
    destruct (length l === length l0); try discriminate.
    case_rewrite (params n0) H; auto.
    destruct (length k === length l); auto.
    apply* SConstr; intros.
    destruct (subtype_infn_constr _ IHh HK R k true p' x k' t1 t2 HU e); auto.
      inversions Ht1. clear -H6. induction* l.
      inversions Ht2. clear -H6. induction* l0.
    destruct (subtype_infn_constr _ IHh HK R k false P x k' t1 t2 HU e); auto.
      inversions Ht1. clear -H6. induction* l.
      inversions Ht2. clear -H6. induction* l0.
  case_rewrite (expansion n) R1.
  destruct p.
  case_rewrite (params n) R2; try discriminate.
  destruct (length k0 === length l); try discriminate.
  destruct (var_freshes (dom K) (length k)) as [vl Fr].
  simpl in HS.
  apply* SSubP.
  set (S := combine (kinds_vars (k ++ k0)) (map Tvar vl ++ l)) in *.
  set (ks := map (kind_map (subst S)) (list_snd k)) in *.
  poses OkK (proj1 HK).
  apply* IHh; clear IHh.
      unfold ks, S; apply* kenv_ok_expansion.
    destruct (expansion_ok1 R1) as [k0' EOk].
    intuition.
    rewrite H in R2; inversions R2; clear R2.
    unfold ks, S; apply* type_subst.
  apply* type_weaken.
Qed.
